"""An Abstract Base Class for collections of signatures, plus implementations.

APIs and functionality
----------------------

Index classes support three sets of API functionality -

'select(...)', which selects subsets of signatures based on ksize, moltype,
and other criteria, including picklists.

'find(...)', and the 'search', 'gather', and 'counter_gather' implementations
built on top of 'find', which search for signatures that match a query.

'signatures()', which yields all signatures in the Index subject to the
selection criteria.

Classes defined in this file
----------------------------

Index - abstract base class for all Index objects.

LinearIndex - simple in-memory storage of signatures.

LazyLinearIndex - lazy selection and linear search of signatures.

ZipFileLinearIndex - simple on-disk storage of signatures.

MultiIndex - in-memory storage and selection of signatures from multiple
index objects, using manifests. All signatures are kept in memory.

StandaloneManifestIndex - load manifests directly, and do lazy loading of
signatures on demand. No signatures are kept in memory.

CounterGather - an ancillary class returned by the 'counter_gather()' method.
"""

import os
import sourmash
from abc import abstractmethod, ABC
from collections import namedtuple, Counter

from sourmash.search import (
    make_jaccard_search_query,
    make_containment_query,
    calc_threshold_from_bp,
)
from sourmash.manifest import CollectionManifest
from sourmash.logging import debug_literal
from sourmash.signature import load_signatures_from_json, save_signatures_to_json

from sourmash.minhash import (
    flatten_and_downsample_scaled,
    flatten_and_downsample_num,
    flatten_and_intersect_scaled,
)

# generic return tuple for Index.search and Index.gather
IndexSearchResult = namedtuple("Result", "score, signature, location")


class Index(ABC):
    # this will be removed soon; see sourmash#1894.
    is_database = False

    # 'manifest', when set, implies efficient selection and direct
    # access to signatures. Signatures may be stored in the manifest
    # or loaded on demand from disk depending on the class, however.
    manifest = None

    @abstractmethod
    def __len__(self):
        "Return the number of signatures in this Index object."

    @property
    def location(self):
        "Return a resolvable location for this index, if possible."
        return None

    @abstractmethod
    def signatures(self):
        "Return an iterator over all signatures in the Index object."

    def signatures_with_location(self):
        "Return an iterator over tuples (signature, location) in the Index."
        for ss in self.signatures():
            yield ss, self.location

    def _signatures_with_internal(self):
        """Return an iterator of tuples (ss, internal_location).

        Unlike 'signatures_with_location()', this iterator should return
        _all_ signatures in the object, not just those that remain after
        selection/filtering.

        This is an internal API for use in generating manifests, and may
        change without warning.

        This method should be implemented separately for each Index object.
        """
        raise NotImplementedError

    @abstractmethod
    def insert(self, signature):
        """ """

    @abstractmethod
    def save(self, path, storage=None, sparseness=0.0, structure_only=False):
        """ """

    @classmethod
    @abstractmethod
    def load(cls, location, leaf_loader=None, storage=None, print_version_warning=True):
        """ """

    def find(self, search_fn, query, **kwargs):
        """Use search_fn to find matching signatures in the index.

        search_fn follows the protocol in JaccardSearch objects.

        Generator. Returns 0 or more IndexSearchResult objects.
        """
        # first: is this query compatible with this search?
        search_fn.check_is_compatible(query)

        # ok! continue!

        # this set of signatures may be heterogenous in scaled/num values;
        # define some processing functions to downsample appropriately.
        query_mh = query.minhash
        assert not query_mh.track_abundance
        if query_mh.scaled:
            # make query and subject compatible w/scaled.
            query_scaled = query_mh.scaled

            def prepare_subject(subj_mh):
                return flatten_and_downsample_scaled(subj_mh, query_scaled)

            def prepare_query(query_mh, subj_mh):
                return flatten_and_downsample_scaled(query_mh, subj_mh.scaled)

        else:  # num
            query_num = query_mh.num

            def prepare_subject(subj_mh):
                return flatten_and_downsample_num(subj_mh, query_num)

            def prepare_query(query_mh, subj_mh):
                return flatten_and_downsample_num(query_mh, subj_mh.num)

        # now, do the search!
        for subj, location in self.signatures_with_location():
            subj_mh = prepare_subject(subj.minhash)
            # note: we run prepare_query here on the original query minhash.
            query_mh = prepare_query(query.minhash, subj_mh)

            assert not query_mh.track_abundance
            assert not subj_mh.track_abundance

            shared_size, total_size = query_mh.intersection_and_union_size(subj_mh)

            query_size = len(query_mh)
            subj_size = len(subj_mh)

            score = search_fn.score_fn(query_size, shared_size, subj_size, total_size)

            if search_fn.passes(score):
                # note: here we yield the original signature, not the
                # downsampled minhash.
                if search_fn.collect(score, subj):
                    yield IndexSearchResult(score, subj, location)

    def search_abund(self, query, *, threshold=None, **kwargs):
        """Return list of IndexSearchResult with angular similarity above 'threshold'.

        Results will be sorted by similarity, highest to lowest.
        """
        if not query.minhash.track_abundance:
            raise TypeError(
                "'search_abund' requires query signature with abundance information"
            )

        # check arguments
        if threshold is None:
            raise TypeError("'search_abund' requires 'threshold'")
        threshold = float(threshold)

        # do the actual search:
        matches = []
        for subj, loc in self.signatures_with_location():
            if not subj.minhash.track_abundance:
                raise TypeError(
                    "'search_abund' requires subject signatures with abundance information"
                )
            score = query.similarity(subj, downsample=True)
            if score >= threshold:
                matches.append(IndexSearchResult(score, subj, loc))

        # sort!
        matches.sort(key=lambda x: -x.score)
        return matches

    def search(
        self,
        query,
        *,
        threshold=None,
        do_containment=False,
        do_max_containment=False,
        best_only=False,
        **kwargs,
    ):
        """Return list of IndexSearchResult with similarity above 'threshold'.

        Results will be sorted by similarity, highest to lowest.

        Optional arguments accepted by all Index subclasses:
          * do_containment: default False. If True, use Jaccard containment.
          * best_only: default False. If True, allow optimizations that
            may. May discard matches better than threshold, but first match
            is guaranteed to be best.
        """
        # check arguments
        if threshold is None:
            raise TypeError("'search' requires 'threshold'")
        threshold = float(threshold)

        search_obj = make_jaccard_search_query(
            do_containment=do_containment,
            do_max_containment=do_max_containment,
            best_only=best_only,
            threshold=threshold,
        )

        # do the actual search:
        matches = list(self.find(search_obj, query, **kwargs))

        # sort!
        matches.sort(key=lambda x: -x.score)
        return matches

    def prefetch(self, query, threshold_bp, **kwargs):
        """Return all matches with minimum overlap.

        Generator. Returns 0 or more IndexSearchResult namedtuples.
        """
        if not self:  # empty database? quit.
            raise ValueError("no signatures to search")

        # default best_only to False
        best_only = kwargs.get("best_only", False)

        search_fn = make_containment_query(
            query.minhash, threshold_bp, best_only=best_only
        )

        yield from self.find(search_fn, query, **kwargs)

    def best_containment(self, query, threshold_bp=None, **kwargs):
        """Return the match with the best Jaccard containment in the Index.

        Returns an IndexSearchResult namedtuple or None.
        """

        results = self.prefetch(query, threshold_bp, best_only=True, **kwargs)
        results = sorted(results, key=lambda x: (-x.score, x.signature.md5sum()))

        try:
            return next(iter(results))
        except StopIteration:
            return None

    def peek(self, query_mh, *, threshold_bp=0):
        """Mimic CounterGather.peek() on top of Index.

        This is implemented for situations where we don't want to use
        'prefetch' functionality. It is a light wrapper around the
        'best_containment(...)' method.
        """
        from sourmash import SourmashSignature

        # build a signature to use with self.gather...
        query_ss = SourmashSignature(query_mh)

        # run query!
        try:
            result = self.best_containment(query_ss, threshold_bp=threshold_bp)
        except ValueError:
            result = None

        if not result:
            return []

        # if matches, calculate intersection & return.
        intersect_mh = flatten_and_intersect_scaled(result.signature.minhash, query_mh)

        return [result, intersect_mh]

    def consume(self, intersect_mh):
        "Mimic CounterGather.consume on top of Index. Yes, this is backwards."
        pass

    def counter_gather(self, query, threshold_bp, **kwargs):
        """Returns an object that permits 'gather' on top of the
        current contents of this Index.

        The default implementation uses `prefetch` underneath, and returns
        the results in a `CounterGather` object. However, alternate
        implementations need only return an object that meets the
        public `CounterGather` interface, of course.
        """
        with query.update() as prefetch_query:
            prefetch_query.minhash = prefetch_query.minhash.flatten()

        # find all matches and construct a CounterGather object.
        counter = CounterGather(prefetch_query)
        for result in self.prefetch(prefetch_query, threshold_bp, **kwargs):
            counter.add(result.signature, location=result.location)

        # tada!
        return counter

    @abstractmethod
    def select(
        self,
        ksize=None,
        moltype=None,
        scaled=None,
        num=None,
        abund=None,
        containment=None,
    ):
        """Return Index containing only signatures that match requirements.

        Current arguments can be any or all of:
        * ksize
        * moltype
        * scaled
        * num
        * containment

        'select' will raise ValueError if the requirements are incompatible
        with the Index subclass.

        'select' may return an empty object or None if no matches can be
        found.
        """


def select_signature(
    ss,
    *,
    ksize=None,
    moltype=None,
    scaled=0,
    num=0,
    containment=False,
    abund=None,
    picklist=None,
):
    "Check that the given signature matches the specified requirements."
    # ksize match?
    if ksize and ksize != ss.minhash.ksize:
        return False

    # moltype match?
    if moltype and moltype != ss.minhash.moltype:
        return False

    # containment requires scaled; similarity does not.
    if containment:
        if not scaled:
            raise ValueError("'containment' requires 'scaled' in Index.select'")
        if not ss.minhash.scaled:
            return False

    # 'scaled' and 'num' are incompatible
    if scaled:
        if ss.minhash.num:
            return False
    if num:
        # note, here we check if 'num' is identical; this can be
        # changed later.
        if ss.minhash.scaled or num != ss.minhash.num:
            return False

    if abund:
        # note: minhash w/abund can always be flattened
        if not ss.minhash.track_abundance:
            return False

    if picklist is not None and ss not in picklist:
        return False

    return True


class LinearIndex(Index):
    """An Index for a collection of signatures. Can load from a .sig file.

    Note: See MultiIndex for an in-memory class that uses manifests.

    Concrete class; signatures held in memory; does not use manifests.
    """

    def __init__(self, _signatures=None, filename=None):
        self._signatures = []
        if _signatures:
            self._signatures = list(_signatures)
        self.filename = filename

    @property
    def location(self):
        return self.filename

    def signatures(self):
        return iter(self._signatures)

    def __bool__(self):
        return bool(self._signatures)

    def __len__(self):
        return len(self._signatures)

    def insert(self, node):
        self._signatures.append(node)

    def save(self, path):
        with open(path, "w") as fp:
            save_signatures_to_json(self.signatures(), fp)

    @classmethod
    def load(cls, location, filename=None):
        "Load signatures from a JSON signature file."
        si = load_signatures_from_json(location, do_raise=True)

        if filename is None:
            filename = location
        lidx = LinearIndex(si, filename=filename)
        return lidx

    def select(self, **kwargs):
        """Return new LinearIndex containing only signatures that match req's.

        Does not raise ValueError, but may return an empty Index.
        """
        _check_select_parameters(**kwargs)

        siglist = []
        for ss in self._signatures:
            if select_signature(ss, **kwargs):
                siglist.append(ss)

        return LinearIndex(siglist, self.location)


class LazyLinearIndex(Index):
    """An Index for lazy linear search of another database.

    Wrapper class; does not use manifests.

    One of the main purposes of this class is to _force_ linear 'find'
    on index objects. So if this class wraps an SBT, for example, the
    SBT find method will be overriden with the linear 'find' from the
    base class. There are very few situations where this is an improvement,
    so use this class wisely!

    A few notes:
    * selection criteria defined by 'select' are only executed when
      signatures are actually requested (hence, 'lazy').
    * this class stores the provided index 'db' in memory. If you need
      a class that does lazy loading of signatures from disk and does not
      store signatures in memory, see StandaloneManifestIndex.
    * if you want efficient manifest-based selection, consider
      MultiIndex (signatures in memory).
    """

    def __init__(self, db, selection_dict={}):
        self.db = db
        self.selection_dict = dict(selection_dict)

    def signatures(self):
        "Return the selected signatures."
        db = self.db.select(**self.selection_dict)
        yield from db.signatures()

    def signatures_with_location(self):
        "Return the selected signatures, with a location."
        db = self.db.select(**self.selection_dict)
        yield from db.signatures_with_location()

    def __bool__(self):
        try:
            next(iter(self.signatures()))
            return True
        except StopIteration:
            return False

    def __len__(self):
        db = self.db.select(**self.selection_dict)
        return len(db)

    def insert(self, node):
        raise NotImplementedError

    def save(self, path):
        raise NotImplementedError

    @classmethod
    def load(cls, path):
        raise NotImplementedError

    def select(self, **kwargs):
        """Return new object yielding only signatures that match req's.

        Does not raise ValueError, but may return an empty Index.
        """
        _check_select_parameters(**kwargs)

        selection_dict = dict(self.selection_dict)
        for k, v in kwargs.items():
            if k in selection_dict:
                if selection_dict[k] != v:
                    raise ValueError(f"cannot select on two different values for {k}")
            selection_dict[k] = v

        return LazyLinearIndex(self.db, selection_dict)


class ZipFileLinearIndex(Index):
    """\
    A read-only collection of signatures in a zip file.

    Does not support `insert` or `save`.

    Concrete class; signatures dynamically loaded from disk; uses manifests.
    """

    is_database = True

    def __init__(
        self,
        storage,
        *,
        selection_dict=None,
        traverse_yield_all=False,
        manifest=None,
        use_manifest=True,
    ):
        self.storage = storage
        self.selection_dict = selection_dict
        self.traverse_yield_all = traverse_yield_all
        self.use_manifest = use_manifest

        # do we have a manifest already? if not, try loading.
        if use_manifest:
            if manifest is not None:
                debug_literal("ZipFileLinearIndex using passed-in manifest")
                self.manifest = manifest
            else:
                self._load_manifest()
        else:
            self.manifest = None

        if self.manifest is not None:
            assert not self.selection_dict, self.selection_dict
        if self.selection_dict:
            assert self.manifest is None

    def _load_manifest(self):
        "Load a manifest if one exists"
        try:
            manifest_data = self.storage.load("SOURMASH-MANIFEST.csv")
        except (KeyError, FileNotFoundError):
            self.manifest = None
        else:
            debug_literal(f"found manifest on load for {self.storage.path}")

            # load manifest!
            from io import StringIO

            manifest_data = manifest_data.decode("utf-8")
            manifest_fp = StringIO(manifest_data)
            self.manifest = CollectionManifest.load_from_csv(manifest_fp)

    def __bool__(self):
        "Are there any matching signatures in this zipfile? Avoid calling len."
        try:
            next(iter(self.signatures()))
        except StopIteration:
            return False

        return True

    def __len__(self):
        "calculate number of signatures."

        # use manifest, if available.
        m = self.manifest
        if self.manifest is not None:
            return len(m)

        # otherwise, iterate across all signatures.
        n = 0
        for _ in self.signatures():
            n += 1
        return n

    @property
    def location(self):
        return self.storage.path

    def insert(self, signature):
        raise NotImplementedError

    def save(self, path):
        raise NotImplementedError

    @classmethod
    def load(cls, location, traverse_yield_all=False, use_manifest=True):
        "Class method to load a zipfile."
        from ..sbt_storage import ZipStorage

        # we can only load from existing zipfiles in this method.
        if not os.path.exists(location):
            raise FileNotFoundError(location)

        storage = ZipStorage(location)
        return cls(
            storage, traverse_yield_all=traverse_yield_all, use_manifest=use_manifest
        )

    def _signatures_with_internal(self):
        """Return an iterator of tuples (ss, internal_location).

        Note: does not limit signatures to subsets.
        """
        # list all the files, without using the Storage interface; currently,
        # 'Storage' does not provide a way to list all the files, so :shrug:.
        for filename in self.storage._filenames():
            # should we load this file? if it ends in .sig OR we are forcing:
            if (
                filename.endswith(".sig")
                or filename.endswith(".sig.gz")
                or self.traverse_yield_all
            ):
                sig_data = self.storage.load(filename)
                for ss in load_signatures_from_json(sig_data):
                    yield ss, filename

    def signatures(self):
        "Load all signatures in the zip file."
        selection_dict = self.selection_dict
        manifest = None
        if self.manifest is not None:
            manifest = self.manifest
            assert not selection_dict

            # yield all signatures found in manifest
            for filename in manifest.locations():
                data = self.storage.load(filename)
                for ss in load_signatures_from_json(data):
                    # in case multiple signatures are in the file, check
                    # to make sure we want to return each one.
                    if ss in manifest:
                        yield ss

        # no manifest! iterate.
        else:
            storage = self.storage
            # if no manifest here, break Storage class encapsulation
            # and go for all the files. (This is necessary to support
            # ad-hoc zipfiles that have no manifests.)
            for filename in storage._filenames():
                # should we load this file? if it ends in .sig OR force:
                if (
                    filename.endswith(".sig")
                    or filename.endswith(".sig.gz")
                    or self.traverse_yield_all
                ):
                    if selection_dict:

                        def select(x):
                            return select_signature(x, **selection_dict)
                    else:

                        def select(x):
                            return True

                    data = self.storage.load(filename)
                    for ss in load_signatures_from_json(data):
                        if select(ss):
                            yield ss

    def select(self, **kwargs):
        "Select signatures in zip file based on ksize/moltype/etc."
        _check_select_parameters(**kwargs)

        # if we have a manifest, run 'select' on the manifest.
        manifest = self.manifest
        traverse_yield_all = self.traverse_yield_all

        if manifest is not None:
            manifest = manifest.select_to_manifest(**kwargs)
            return ZipFileLinearIndex(
                self.storage,
                selection_dict=None,
                traverse_yield_all=traverse_yield_all,
                manifest=manifest,
                use_manifest=True,
            )
        else:
            # no manifest? just pass along all the selection kwargs to
            # the new ZipFileLinearIndex.

            assert manifest is None
            if self.selection_dict:
                # combine selects...
                d = dict(self.selection_dict)
                for k, v in kwargs.items():
                    if k in d:
                        if d[k] is not None and d[k] != v:
                            raise ValueError(f"incompatible select on '{k}'")
                    d[k] = v
                kwargs = d

            return ZipFileLinearIndex(
                self.storage,
                selection_dict=kwargs,
                traverse_yield_all=traverse_yield_all,
                manifest=None,
                use_manifest=False,
            )


class CounterGather:
    """This is an ancillary class that is used to implement "fast
    gather", post-prefetch. It tracks and summarize matches for
    efficient min-set-cov/'gather'.

    The class constructor takes a query MinHash that must be scaled, and
    then takes signatures that have overlaps with the query (via 'add').

    After all overlapping signatures have been loaded, the 'peek'
    method is then used at each stage of the 'gather' procedure to
    find the best match, and the 'consume' method is used to remove
    a match from this counter.

    This particular implementation maintains a collections.Counter that
    is used to quickly find the best match when 'peek' is called, but
    other implementations are possible ;).

    Note that redundant matches (SourmashSignature objects) with
    duplicate md5s are collapsed inside the class, because we use the
    md5sum as a key into the dictionary used to store matches.
    """

    def __init__(self, query):
        "Constructor - takes a query SourmashSignature."
        query_mh = query.minhash
        if not query_mh.scaled:
            raise ValueError("gather requires scaled signatures")

        # track query
        self.orig_query_mh = query_mh.copy().flatten()
        self.scaled = query_mh.scaled

        # use these to track loaded matches & their locations
        self.siglist = {}
        self.locations = {}

        # ...and also track overlaps with the progressive query
        self.counter = Counter()

        # fence to make sure we do add matches once query has started.
        self.query_started = 0

    def add(self, ss, *, location=None, require_overlap=True):
        "Add this signature in as a potential match."
        if self.query_started:
            raise ValueError("cannot add more signatures to counter after peek/consume")

        # upon insertion, count & track overlap with the specific query.
        overlap = self.orig_query_mh.count_common(ss.minhash, True)
        if overlap:
            md5 = ss.md5sum()

            self.counter[md5] = overlap
            self.siglist[md5] = ss
            self.locations[md5] = location

            # note: scaled will be max of all matches.
            self.downsample(ss.minhash.scaled)
        elif require_overlap:
            raise ValueError("no overlap between query and signature!?")

    def downsample(self, scaled):
        "Track highest scaled across all possible matches."
        if scaled > self.scaled:
            self.scaled = scaled
        return self.scaled

    def signatures(self):
        "Return all signatures."
        yield from self.siglist.values()

    @property
    def union_found(self):
        """Return a MinHash containing all found hashes in the query.

        This calculates the union of the found matches, intersected
        with the original query.
        """
        orig_query_mh = self.orig_query_mh

        # create empty MinHash from orig query
        found_mh = orig_query_mh.copy_and_clear()

        # for each match, intersect match with query & then add to found_mh.
        for ss in self.siglist.values():
            intersect_mh = flatten_and_intersect_scaled(ss.minhash, orig_query_mh)
            found_mh.add_many(intersect_mh)

        return found_mh

    def peek(self, cur_query_mh, *, threshold_bp=0):
        "Get next 'gather' result for this database, w/o changing counters."
        self.query_started = 1

        # empty? nothing to search.
        counter = self.counter
        if not counter:
            return []

        siglist = self.siglist
        assert siglist

        scaled = self.downsample(cur_query_mh.scaled)
        cur_query_mh = cur_query_mh.downsample(scaled=scaled)

        if not cur_query_mh:  # empty query? quit.
            return []

        # CTB: could probably remove this check unless debug requested.
        if cur_query_mh.contained_by(self.orig_query_mh, downsample=True) < 1:
            raise ValueError("current query not a subset of original query")

        # are we setting a threshold?
        try:
            x = calc_threshold_from_bp(threshold_bp, scaled, len(cur_query_mh))
            threshold, n_threshold_hashes = x
        except ValueError:
            # too high to ever match => exit
            return []

        # Find the best match using the internal Counter.
        most_common = counter.most_common()
        dataset_id, match_size = most_common[0]

        # below threshold? no match!
        if match_size < n_threshold_hashes:
            return []

        ## at this point, we have a legitimate match above threshold!

        # pull match and location.
        match = siglist[dataset_id]

        # calculate containment
        # CTB: this check is probably redundant with intersect_mh calc, below.
        cont = cur_query_mh.contained_by(match.minhash, downsample=True)
        assert cont
        assert cont >= threshold

        # calculate intersection of this "best match" with query.
        match_mh = match.minhash.downsample(scaled=scaled).flatten()
        intersect_mh = cur_query_mh & match_mh
        location = self.locations[dataset_id]

        # build result & return intersection
        return (IndexSearchResult(cont, match, location), intersect_mh)

    def consume(self, intersect_mh):
        "Maintain the internal counter by removing the given hashes."
        self.query_started = 1

        if not intersect_mh:
            return

        siglist = self.siglist
        counter = self.counter

        most_common = counter.most_common()

        # Prepare counter for finding the next match by decrementing
        # all hashes found in the current match in other datasets;
        # remove empty datasets from counter, too.
        for dataset_id, _ in most_common:
            # CTB: note, remaining_mh may not be at correct scaled here.
            # this means that counters that _should_ be empty might not
            # _be_ empty in some situations.  This does not
            # lead to incorrect results, merely potentially overfull
            # 'counter' objects. The tradeoffs to fixing this would
            # need to be examined! (This could be fixed in self.downsample().)
            remaining_mh = siglist[dataset_id].minhash
            intersect_count = intersect_mh.count_common(remaining_mh, downsample=True)
            if intersect_count:
                counter[dataset_id] -= intersect_count
                if counter[dataset_id] == 0:
                    del counter[dataset_id]


class MultiIndex(Index):
    """
    Load a collection of signatures, and retain their original locations.

    One specific use for this is when loading signatures from a directory;
    MultiIndex will record which specific files provided which
    signatures.

    Creates a manifest on load.

    Note: this is an in-memory collection, and does not do lazy loading:
    all signatures are loaded upon instantiation and kept in memory.

    There are a variety of loading functions:
    * `load` takes a list of already-loaded Index objects,
      together with a list of their locations.
    * `load_from_directory` traverses a directory to load files within.
    * `load_from_path` takes an arbitrary pathname and tries to load it
      as a directory, or as a .sig file.
    * `load_from_pathlist` takes a text file full of pathnames and tries
      to load them all.

    Concrete class; signatures held in memory; builds and uses manifests.
    """

    def __init__(self, manifest, parent, *, prepend_location=False):
        """Constructor; takes manifest containing signatures, together with
        the top-level location.
        """
        self.manifest = manifest
        self.parent = parent
        self.prepend_location = prepend_location

        if prepend_location and self.parent is None:
            raise ValueError("must set 'parent' if 'prepend_location' is set")

    @property
    def location(self):
        return self.parent

    def signatures(self):
        for row in self.manifest.rows:
            yield row["signature"]

    def signatures_with_location(self):
        for row in self.manifest.rows:
            loc = row["internal_location"]
            # here, 'parent' may have been removed from internal_location
            # for directories; if so, add it back in.
            if self.prepend_location:
                loc = os.path.join(self.parent, loc)
            yield row["signature"], loc

    def _signatures_with_internal(self):
        """Return an iterator of tuples (ss, location)

        CTB note: here, 'internal_location' is the source file for the
        index. This is a special feature of this (in memory) class.
        """
        for row in self.manifest.rows:
            yield row["signature"], row["internal_location"]

    def __len__(self):
        if self.manifest is None:
            return 0

        return len(self.manifest)

    def insert(self, *args):
        raise NotImplementedError

    @classmethod
    def load(cls, index_list, source_list, parent, *, prepend_location=False):
        """Create a MultiIndex from already-loaded indices.

        Takes two arguments: a list of Index objects, and a matching list
        of source strings (filenames, etc.)  If the source is not None,
        then it will be used to override the location provided by the
        matching Index object.
        """
        assert len(index_list) == len(source_list)

        # yield all signatures + locations
        def sigloc_iter():
            for idx, iloc in zip(index_list, source_list):
                # override internal location if location is explicitly provided
                if iloc is None:
                    iloc = idx.location
                for ss in idx.signatures():
                    yield ss, iloc

        # build manifest; note, ALL signatures are stored in memory.
        # CTB: could do this on demand?
        # CTB: should we use get_manifest functionality?
        # CTB: note here that the manifest is created by iteration
        # *even if it already exists.* This could be changed to be more
        # efficient... but for now, use StandaloneManifestIndex if you
        # want to avoid this when loading from multiple files.
        manifest = CollectionManifest.create_manifest(sigloc_iter())

        # create!
        return cls(manifest, parent, prepend_location=prepend_location)

    @classmethod
    def load_from_directory(cls, pathname, *, force=False):
        """Create a MultiIndex from a directory.

        Takes directory path plus optional boolean 'force'. Attempts to
        load all files ending in .sig or .sig.gz, by default; if 'force' is
        True, will attempt to load _all_ files, ignoring errors.

        Will not load anything other than JSON signature files.
        """
        from ..sourmash_args import traverse_find_sigs

        if not os.path.isdir(pathname):
            raise ValueError(f"'{pathname}' must be a directory.")

        index_list = []
        source_list = []

        traversal = traverse_find_sigs([pathname], yield_all_files=force)
        for thisfile in traversal:
            try:
                idx = LinearIndex.load(thisfile)
                index_list.append(idx)

                rel = os.path.relpath(thisfile, pathname)
                source_list.append(rel)
            except (OSError, sourmash.exceptions.SourmashError) as exc:
                if force:
                    continue  # ignore error
                else:
                    raise ValueError(exc)  # stop loading!

        # did we load anything? if not, error
        if not index_list:
            raise ValueError(f"no signatures to load under directory '{pathname}'")

        return cls.load(index_list, source_list, pathname, prepend_location=True)

    @classmethod
    def load_from_path(cls, pathname, force=False):
        """
        Create a MultiIndex from a path (filename or directory).

        Note: this only uses LinearIndex.load(...), so will only load
        signature JSON files.
        """
        if not os.path.exists(pathname):
            raise ValueError(f"'{pathname}' must exist.")

        if os.path.isdir(pathname):  # traverse
            return cls.load_from_directory(pathname, force=force)

        # load as a .sig/JSON file
        index_list = []
        source_list = []
        try:
            idx = LinearIndex.load(pathname)
            index_list = [idx]
            source_list = [pathname]
        except (OSError, sourmash.exceptions.SourmashError):
            if not force:
                raise ValueError(f"no signatures to load from '{pathname}'")
            return None

        return cls.load(index_list, source_list, pathname)

    @classmethod
    def load_from_pathlist(cls, filename):
        """Create a MultiIndex from all files listed in a text file.

        Note: this will attempt to load signatures from each file,
        including zip collections, etc; it uses 'load_file_as_index'
        underneath.
        """
        from ..sourmash_args import load_pathlist_from_file, load_file_as_index

        idx_list = []
        src_list = []

        file_list = load_pathlist_from_file(filename)
        for fname in file_list:
            idx = load_file_as_index(fname)
            src = fname

            idx_list.append(idx)
            src_list.append(src)

        return cls.load(idx_list, src_list, filename)

    def save(self, *args):
        raise NotImplementedError

    def select(self, **kwargs):
        "Run 'select' on the manifest."
        _check_select_parameters(**kwargs)
        new_manifest = self.manifest.select_to_manifest(**kwargs)
        return MultiIndex(
            new_manifest, self.parent, prepend_location=self.prepend_location
        )


class StandaloneManifestIndex(Index):
    """Load a standalone manifest as an Index.

    This class is useful for the situation where you have a directory
    with many signature collections underneath it, and you don't want to load
    every collection each time you run sourmash.

    Instead, you can run 'sourmash sig collect <directory> -o <manifest>' to
    output a manifest and then use this class to load <manifest> directly.
    Sketch type selection, picklists, and pattern matching will all work
    directly on the manifest and will load signatures only upon demand.

    One feature of this class is that external paths to sketches in
    the 'internal_location' field of the manifests will be loaded
    properly.  This permits manifests to be constructed for various
    collections of signatures that reside elsewhere, and not just
    below a single directory prefix. By default paths are interpreted
    relative to the location of the manifest, unless an absolute path
    is provided in the 'internal_location' field.

    StandaloneManifestIndex does _not_ store signatures in memory.

    This class also overlaps in concept with MultiIndex when
    MultiIndex.load_from_pathlist is used to load other Index
    objects. However, this class does not store any signatures in
    memory, unlike MultiIndex.

    """

    is_database = True

    def __init__(self, manifest, location, *, prefix=None):
        """Create object. 'location' is path of manifest file, 'prefix' is
        prepended to signature paths when loading non-abspaths."""
        assert manifest is not None
        self.manifest = manifest
        self._location = location
        self.prefix = prefix

    @classmethod
    def load(cls, location, *, prefix=None):
        """Load manifest file from given location.

        If prefix is None (default), it is automatically set from dirname.
        Set prefix='' to avoid this, or provide an explicit prefix.
        """
        if not os.path.isfile(location):
            raise ValueError(f"provided manifest location '{location}' is not a file")

        m = CollectionManifest.load_from_filename(location)

        if prefix is None:
            # by default, calculate paths relative to manifest location.
            prefix = os.path.dirname(location)

        return cls(m, location, prefix=prefix)

    @property
    def location(self):
        "Return the path to this manifest."
        return self._location

    def signatures_with_location(self):
        "Return an iterator over all signatures and their locations."
        yield from self._signatures_with_internal()

    def signatures(self):
        "Return an iterator over all signatures."
        for ss, loc in self._signatures_with_internal():
            yield ss

    def _signatures_with_internal(self):
        """Return an iterator over all sigs of (sig, internal_location)

        Note that this is implemented differently from most Index
        objects in that it only lists subselected parts of the
        manifest, and not the original manifest. This was done out of
        convenience: we don't currently have access to the original
        manifest in this class.
        """
        # collect all internal locations
        picklist = self.manifest.to_picklist()
        for iloc in self.manifest.locations():
            # prepend location with prefix?
            if not iloc.startswith("/") and self.prefix:
                iloc = os.path.join(self.prefix, iloc)

            idx = sourmash.load_file_as_index(iloc)
            idx = idx.select(picklist=picklist)
            for ss in idx.signatures():
                yield ss, iloc

    def __len__(self):
        "Number of signatures in this manifest (after any select)."
        return len(self.manifest)

    def __bool__(self):
        "Is this manifest empty?"
        return bool(self.manifest)

    def save(self, *args):
        raise NotImplementedError

    def insert(self, *args):
        raise NotImplementedError

    def select(self, **kwargs):
        "Run 'select' on the manifest."
        _check_select_parameters(**kwargs)
        new_manifest = self.manifest.select_to_manifest(**kwargs)
        return StandaloneManifestIndex(new_manifest, self._location, prefix=self.prefix)


def _check_select_parameters(**kw):
    "Check 'select' parameters for types/conversion."
    params = set(kw)
    params -= {"ksize", "num", "moltype", "scaled", "abund", "picklist", "containment"}
    if params:
        raise ValueError(f"unknown 'select' parameters: {params}")

    ksize = kw.get("ksize")
    if ksize is not None:
        if not isinstance(ksize, int):
            raise ValueError(
                f"ksize value '{ksize}' must be an integer, is: {type(ksize)}"
            )

    moltype = kw.get("moltype")
    if moltype is not None:
        if moltype not in ["DNA", "protein", "dayhoff", "hp", "skipm1n3", "skipm2n3"]:
            raise ValueError(f"unknown moltype: {moltype}")

    scaled = kw.get("scaled")
    if scaled is not None:
        if not isinstance(scaled, int):
            raise ValueError(
                f"scaled value '{scaled}' must be an integer, is: {type(scaled)}"
            )

    containment = kw.get("containment")
    if containment is not None:
        if not isinstance(containment, bool):
            raise ValueError(
                f"containment value '{containment}' must be a bool, is: {type(containment)}"
            )

    abund = kw.get("abund")
    if abund is not None:
        if not isinstance(abund, bool):
            raise ValueError(f"abund value '{abund}' must be a bool, is: {type(abund)}")

    num = kw.get("num")
    if num is not None:
        if not isinstance(num, int):
            raise ValueError(f"num value '{num}' must be an integer, is: {type(num)}")



"""show k-mers/sequences matching the signature hashes"""

usage = """

### `sourmash signature kmers` - extract k-mers and/or sequences that match to signatures

Given one or more compatible sketches and some sequence files, extract
the k-mers and/or sequences corresponding to the hash values in the
sketch. Because the sourmash hash function is one-way, this requires
FASTA or FASTQ sequence files in addition to the sketch.

For example,

sourmash sig kmers --signatures sig1.sig --sequences seqfile.fasta \
    --save-sequences matches.fasta --save-kmers kmer-matches.csv

will search `seqfile.fasta` for matching sequences and k-mers,
and produce two files. The file `matches.fasta` will contain FASTA
sequences that match the hashes in the input signature, while the
file `kmer-matches.csv` provides the matching k-mers and hash values,
together with their originating filename and sequence name.

If the sketch is a protein sketch (protein, dayhoff, or hp), then
the input sequences are assumed to be protein. To search DNA sequences
for translated protein hashes, provide the `--translate` flag to `sig kmers`.

`--save-sequences` and `--save-kmers` are both optional.  If neither are
given, basic statistics on k-mer matching are given.

Please note that `--save-kmers` can be very slow on large files!

The input sketches are the source of the input hashes.  So, for example,
If `--scaled=1` sketches are provided, `sig kmers` can be used to
yield all the k-mers and their matching hashes.  Likewise, if the
sketch is built from the intersection of two other sketches, only
the k-mers and hash values present in both sketches will be used.

Likewise, the input sequences are used for matching; they do not need
to be the same sequences that were used to create the sketches.
Input sequences can be in FASTA or FASTQ format, and either flat text
or compressed with gzip or bzip2; formats are auto-detected.

By default, `sig kmers` ignores bad k-mers (e.g. non-ACGT characters
in DNA). If `--check-sequence` is provided, `sig kmers` will error
exit on the first bad k-mer.  If `--check-sequence --force` is provided,
`sig kmers` will provide error messages (and skip bad sequences), but
will continue processing input sequences.

"""

from sourmash.cli.utils import add_moltype_args, add_ksize_arg, add_picklist_args


def subparser(subparsers):
    subparser = subparsers.add_parser("kmers", description=__doc__, usage=usage)
    subparser.add_argument("--signatures", nargs="*", default=[])
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)

    subparser.add_argument(
        "--sequences",
        nargs="+",
        required=True,
        help="FASTA/FASTQ/bz2/gz files with sequences",
    )

    subparser.add_argument(
        "--save-kmers", help="save k-mers and hash values to a CSV file"
    )
    subparser.add_argument(
        "--save-sequences", help="save sequences with matching hashes to a FASTA file"
    )
    subparser.add_argument(
        "--translate",
        action="store_true",
        help="translate DNA k-mers into amino acids (for protein, dayhoff, and hp sketches)",
    )
    subparser.add_argument(
        "--check-sequence",
        action="store_true",
        help="complain if input sequence is invalid (NOTE: only checks DNA)",
    )


def main(args):
    import sourmash

    return sourmash.sig.__main__.kmers(args)



"""
Functions implementing the 'sketch' subcommands and related functions.
"""

import sys
import os
from collections import defaultdict, Counter
import csv
import shlex

import screed

import sourmash
from .signature import SourmashSignature
from .logging import notify, error, set_quiet, print_results
from sourmash import sourmash_args
from sourmash.sourmash_args import check_scaled_bounds, check_num_bounds
from sourmash.sig.__main__ import _summarize_manifest, _SketchInfo
from sourmash.manifest import CollectionManifest
from .utils import RustObject
from ._lowlevel import ffi, lib

DEFAULT_MMHASH_SEED = 42

DEFAULTS = dict(
    dna="k=31,scaled=1000,noabund",
    protein="k=10,scaled=200,noabund",
    dayhoff="k=16,scaled=200,noabund",
    hp="k=42,scaled=200,noabund",
)


def _parse_params_str(params_str):
    "Parse a parameter string of the form 'k=ks,num=num,scaled=scaled,abund'."
    moltype = None
    params = {}
    params["ksize"] = []
    items = params_str.split(",")
    for item in items:
        if item == "abund":
            params["track_abundance"] = True
        elif item == "noabund":
            params["track_abundance"] = False
        elif item.startswith("k"):
            if len(item) < 3 or item[1] != "=":
                raise ValueError("k takes a parameter, e.g. 'k=31'")
            params["ksize"].append(int(item[2:]))
        elif item.startswith("num"):
            if len(item) < 5 or item[3] != "=":
                raise ValueError("num takes a parameter, e.g. 'num=500'")
            if params.get("scaled"):
                raise ValueError("cannot set both num and scaled in a single minhash")
            try:
                num = item[4:]
                num = int(num)
            except ValueError:
                raise ValueError(f"cannot parse num='{num}' as a number")

            num = check_num_bounds(num)

            params["num"] = int(item[4:])
            params["scaled"] = 0
        elif item.startswith("scaled"):
            if len(item) < 8 or item[6] != "=":
                raise ValueError("scaled takes a parameter, e.g. 'scaled=1000'")
            if params.get("num"):
                raise ValueError("cannot set both num and scaled in a single minhash")
            try:
                scaled = item[7:]
                scaled = int(scaled)
            except ValueError:
                raise ValueError(f"cannot parse scaled='{scaled}' as an integer")

            scaled = check_scaled_bounds(scaled)

            params["scaled"] = scaled
            params["num"] = 0
        elif item.startswith("seed"):
            if len(item) < 6 or item[4] != "=":
                raise ValueError("seed takes a parameter, e.g. 'seed=42'")
            params["seed"] = int(item[5:])
        elif item in ("protein", "dayhoff", "hp", "dna"):
            moltype = item
        else:
            raise ValueError(f"unknown component '{item}' in params string")

    return moltype, params


class _signatures_for_sketch_factory:
    "Build sigs on demand, based on args input to 'sketch'."

    def __init__(self, params_str_list, default_moltype):
        # first, set up defaults per-moltype
        defaults = {}
        for moltype, pstr in DEFAULTS.items():
            mt, d = _parse_params_str(pstr)
            assert mt is None  # defaults cannot have moltype set!
            defaults[moltype] = d
        self.defaults = defaults

        # next, fill out params_list
        self.params_list = []
        self.mult_ksize_by_3 = True

        if params_str_list:
            # parse each params_str passed in, using default_moltype if none
            # provided.
            for params_str in params_str_list:
                moltype, params = _parse_params_str(params_str)
                if moltype and moltype != "dna" and default_moltype == "dna":
                    raise ValueError(
                        f"Incompatible sketch type ({default_moltype}) and parameter override ({moltype}) in '{params_str}'; maybe use 'sketch translate'?"
                    )
                elif moltype == "dna" and default_moltype and default_moltype != "dna":
                    raise ValueError(
                        f"Incompatible sketch type ({default_moltype}) and parameter override ({moltype}) in '{params_str}'"
                    )
                elif moltype is None:
                    if default_moltype is None:
                        raise ValueError(
                            "No default moltype and none specified in param string"
                        )
                    moltype = default_moltype

                self.params_list.append((moltype, params))
        else:
            if default_moltype is None:
                raise ValueError(
                    "No default moltype and none specified in param string"
                )
            # no params str? default to a single sig, using default_moltype.
            self.params_list.append((default_moltype, {}))

    def get_compute_params(self, *, split_ksizes=False):
        for moltype, params_d in self.params_list:
            # get defaults for this moltype from self.defaults:
            default_params = self.defaults[moltype]
            def_seed = default_params.get("seed", DEFAULT_MMHASH_SEED)
            def_num = default_params.get("num", 0)
            def_abund = default_params["track_abundance"]
            def_scaled = default_params.get("scaled", 0)
            def_dna = default_params.get("is_dna", moltype == "dna")
            def_protein = default_params.get("is_protein", moltype == "protein")
            def_dayhoff = default_params.get("is_dayhoff", moltype == "dayhoff")
            def_hp = default_params.get("is_hp", moltype == "hp")

            # handle ksize specially, for now - multiply by three?
            def_ksizes = default_params["ksize"]
            ksizes = params_d.get("ksize")
            if not ksizes:
                ksizes = def_ksizes

            # 'command sketch' adjusts k-mer sizes by 3 if non-DNA sketch.
            if self.mult_ksize_by_3 and not def_dna:
                ksizes = [k * 3 for k in ksizes]

            def make_param(ksizes):
                return ComputeParameters(
                    ksizes=ksizes,
                    seed=params_d.get("seed", def_seed),
                    protein=def_protein,
                    dayhoff=def_dayhoff,
                    hp=def_hp,
                    dna=def_dna,
                    num_hashes=params_d.get("num", def_num),
                    track_abundance=params_d.get("track_abundance", def_abund),
                    scaled=params_d.get("scaled", def_scaled),
                )

            if split_ksizes:
                for ksize in ksizes:
                    params_obj = make_param([ksize])
                    yield params_obj
            else:
                params_obj = make_param(ksizes)
                yield params_obj

    def __call__(self, *, split_ksizes=False):
        "Produce a new set of signatures built to match the param strings."
        sigs = []
        for params in self.get_compute_params(split_ksizes=split_ksizes):
            sig = SourmashSignature.from_params(params)
            sigs.append(sig)

        return sigs


def _add_from_file_to_filenames(args):
    "Add filenames from --from-file to args.filenames"
    from .sourmash_args import load_pathlist_from_file

    if args.from_file:
        file_list = load_pathlist_from_file(args.from_file)
        args.filenames.extend(file_list)


def _execute_sketch(args, signatures_factory):
    "Once configured, run 'sketch' the same way underneath."
    set_quiet(args.quiet)

    if not args.filenames:
        error("error: no input filenames provided! nothing to do - exiting.")
        sys.exit(-1)

    if args.license != "CC0":
        error("error: sourmash only supports CC0-licensed signatures. sorry!")
        sys.exit(-1)

    notify(f"computing signatures for files: {', '.join(args.filenames)}")

    if args.merge and not args.output:
        error("ERROR: must specify -o with --merge")
        sys.exit(-1)

    if args.output and args.output_dir:
        error("ERROR: --output-dir doesn't make sense with -o/--output")
        sys.exit(-1)

    # get number of output sigs:
    num_sigs = len(signatures_factory.params_list)
    notify(f"Computing a total of {num_sigs} signature(s) for each input.")

    if num_sigs == 0:
        error("...nothing to calculate!? Exiting!")
        sys.exit(-1)

    if args.merge:  # single name specified - combine all
        _compute_merged(args, signatures_factory)
    else:  # compute individual signatures
        _compute_individual(args, signatures_factory)


def dna(args):
    """Compute a DNA signature for one or more files.

    CTB: make usable via Python?
    """
    # for dna:
    args.input_is_protein = False

    try:
        signatures_factory = _signatures_for_sketch_factory(args.param_string, "dna")
    except ValueError as e:
        error(f"Error creating signatures: {str(e)}")
        sys.exit(-1)

    _add_from_file_to_filenames(args)
    _execute_sketch(args, signatures_factory)


def protein(args):
    """Compute a protein signature for one or more files.

    CTB: make usable via Python?
    """
    # for protein:
    args.input_is_protein = True
    args.check_sequence = False

    # provide good defaults for dayhoff/hp/protein!
    if args.dayhoff and args.hp:
        raise ValueError("cannot set both --dayhoff and --hp")
    if args.dayhoff:
        moltype = "dayhoff"
    elif args.hp:
        moltype = "hp"
    else:
        moltype = "protein"

    try:
        signatures_factory = _signatures_for_sketch_factory(args.param_string, moltype)
    except ValueError as e:
        error(f"Error creating signatures: {str(e)}")
        sys.exit(-1)

    _add_from_file_to_filenames(args)
    _execute_sketch(args, signatures_factory)


def translate(args):
    """Compute protein signatures from DNA/RNA, for one or more files.

    CTB: make usable via Python?
    """
    # for translate:
    args.input_is_protein = False

    # provide good defaults for dayhoff/hp/protein!
    if args.dayhoff and args.hp:
        raise ValueError("cannot set both --dayhoff and --hp")
    if args.dayhoff:
        moltype = "dayhoff"
    elif args.hp:
        moltype = "hp"
    else:
        moltype = "protein"

    try:
        signatures_factory = _signatures_for_sketch_factory(args.param_string, moltype)
    except ValueError as e:
        error(f"Error creating signatures: {str(e)}")
        sys.exit(-1)

    _add_from_file_to_filenames(args)
    _execute_sketch(args, signatures_factory)


def _compute_sigs(to_build, output, *, check_sequence=False):
    "actually build the signatures in 'to_build' and output them to 'output'"
    save_sigs = sourmash_args.SaveSignaturesToLocation(output)
    save_sigs.open()

    for (name, filename), param_objs in to_build.items():
        assert param_objs

        # now, set up to iterate over sequences.
        with screed.open(filename) as screed_iter:
            if not screed_iter:
                error(f"ERROR: no sequences found in '{filename}'?!")
                sys.exit(-1)

            # build the set of empty sigs
            sigs = []

            is_dna = param_objs[0].dna
            for p in param_objs:
                if p.dna:
                    assert is_dna
                sig = SourmashSignature.from_params(p)
                sigs.append(sig)

            input_is_protein = not is_dna

            # read sequence records & sketch
            notify(f"... reading sequences from {filename}")
            for n, record in enumerate(screed_iter):
                if n % 10000 == 0:
                    if n:
                        notify("\r...{} {}", filename, n, end="")

                try:
                    add_seq(sigs, record.sequence, input_is_protein, check_sequence)
                except ValueError as exc:
                    error(f"ERROR when reading from '{filename}' - ")
                    error(str(exc))
                    sys.exit(-1)

            notify("...{} {} sequences", filename, n, end="")

            set_sig_name(sigs, filename, name)
            for sig in sigs:
                save_sigs.add(sig)

            notify(
                f"calculated {len(sigs)} signatures for {n + 1} sequences in {filename}"
            )

    save_sigs.close()
    notify(
        f"saved {len(save_sigs)} signature(s) to '{save_sigs.location}'. Note: signature license is CC0."
    )


def _output_csv_info(filename, sigs_to_build):
    "output information about what signatures to build, in CSV format"
    output_n = 0
    with sourmash_args.FileOutputCSV(filename) as csv_fp:
        w = csv.DictWriter(
            csv_fp,
            fieldnames=["filename", "sketchtype", "output_index", "name", "param_strs"],
        )
        w.writeheader()

        output_n = 0
        for (name, filename), param_objs in sigs_to_build.items():
            param_strs = []

            # should all be the same!
            if param_objs[0].dna:
                assert all(p.dna for p in param_objs)
                sketchtype = "dna"
            else:
                assert not any(p.dna for p in param_objs)
                sketchtype = "protein"

            for p in param_objs:
                param_strs.append(p.to_param_str())

            row = dict(
                filename=filename,
                sketchtype=sketchtype,
                param_strs="-p " + " -p ".join(param_strs),
                name=name,
                output_index=output_n,
            )

            w.writerow(row)

            output_n += 1


def fromfile(args):
    if args.license != "CC0":
        error("error: sourmash only supports CC0-licensed signatures. sorry!")
        sys.exit(-1)

    if args.output_signatures and os.path.exists(args.output_signatures):
        if not args.force_output_already_exists:
            error(
                f"** ERROR: output location '{args.output_signatures}' already exists!"
            )
            error("** Not overwriting/appending.")
            error(
                "** Use --force-output-already-exists if you want to overwrite/append."
            )
            sys.exit(-1)

    # now, create the set of desired sketch specs.
    try:
        # omit a default moltype - must be provided in param string.
        sig_factory = _signatures_for_sketch_factory(args.param_string, None)
    except ValueError as e:
        error(f"Error creating signatures: {str(e)}")
        sys.exit(-1)

    # take the signatures factory => convert into a bunch of ComputeParameters
    # objects.
    build_params = list(sig_factory.get_compute_params(split_ksizes=True))

    # confirm that they do not adjust seed, which is not supported in
    # 'fromfile' b/c we don't store that info in manifests. (see #1849)
    for p in build_params:
        if p.seed != DEFAULT_MMHASH_SEED:
            error("** ERROR: cannot set 'seed' in 'sketch fromfile'")
            sys.exit(-1)

    # cross-product all of the names in the input CSV file
    # with the sketch spec(s) provided on the command line.

    to_build = defaultdict(list)
    all_names = {}
    total_rows = 0
    skipped_sigs = 0
    n_missing_name = 0
    n_duplicate_name = 0
    duplicate_names = set()

    for csvfile in args.csvs:
        with sourmash_args.FileInputCSV(csvfile) as r:
            for row in r:
                name = row["name"]
                if not name:
                    n_missing_name += 1
                    continue

                genome = row["genome_filename"]
                proteome = row["protein_filename"]
                total_rows += 1

                if name in all_names:
                    n_duplicate_name += 1
                    duplicate_names.add(name)
                else:
                    all_names[name] = (genome, proteome)

    fail_exit = False
    if n_duplicate_name:
        if args.report_duplicated:
            notify("duplicated:\n" + "\n".join(sorted(duplicate_names)))
        error(
            f"** ERROR: {n_duplicate_name} entries have duplicate 'name' records. Exiting!"
        )
        fail_exit = True

    if n_missing_name:
        error(f"** ERROR: {n_missing_name} entries have blank 'name's? Exiting!")
        fail_exit = True

    if fail_exit:
        sys.exit(-1)

    # load manifests from '--already-done' databases => turn into
    # ComputeParameters objects, indexed by name.

    already_done = defaultdict(list)
    already_done_rows = []
    for filename in args.already_done:
        idx = sourmash.load_file_as_index(filename)
        manifest = idx.manifest
        assert manifest

        # for each manifest row,
        for row in manifest.rows:
            name = row["name"]
            if name:
                # build a ComputeParameters object for later comparison
                p = ComputeParameters.from_manifest_row(row)

                # add to list for this name
                already_done[name].append(p)

                # matching name? check if we already have sig. if so, store!
                if name in all_names:
                    if p in build_params:
                        already_done_rows.append(row)

    already_done_manifest = CollectionManifest(already_done_rows)
    if args.already_done:
        notify(f"Loaded {len(already_done)} pre-existing names from manifest(s)")
        notify(f"collected {len(already_done_rows)} rows for already-done signatures.")

    ## now check which are already done and track only those that are
    ## need to be done.

    total_sigs = 0
    missing = defaultdict(list)
    missing_count = 0
    for name, (genome, proteome) in all_names.items():
        plist = already_done.get(name, [])

        # check list of already done against build parameters
        for p in build_params:
            total_sigs += 1

            # does this signature already exist?
            if p not in plist:
                # nope - figure out genome/proteome needed
                filename = genome if p.dna else proteome
                filetype = "genome" if p.dna else "proteome"

                if filename:
                    # add to build list
                    to_build[(name, filename)].append(p)
                else:
                    notify(f"WARNING: fromfile entry '{name}' is missing a {filetype}")
                    missing[name].append(p)
                    missing_count += 1
            else:
                skipped_sigs += 1

    ## we now have 'to_build' which contains the things we can build,
    ## and 'missing', which contains anything we cannot build. Report!

    notify(f"Read {total_rows} rows, requesting that {total_sigs} signatures be built.")

    if already_done_manifest:
        info_d = _summarize_manifest(already_done_manifest)
        print_results("---")
        print_results("summary of already-done sketches:")

        for ski in info_d["sketch_info"]:
            mh_type = f"num={ski['num']}" if ski["num"] else f"scaled={ski['scaled']}"
            mh_abund = ", abund" if ski["abund"] else ""

            sketch_str = f"{ski['count']} sketches with {ski['moltype']}, k={ski['ksize']}, {mh_type}{mh_abund}"

            print_results(f"   {sketch_str: <50} {ski['n_hashes']} total hashes")

        print_results("---")

    if args.output_manifest_matching:
        already_done_manifest.write_to_filename(args.output_manifest_matching)
        notify(
            f"output {len(already_done_manifest)} already-done signatures to '{args.output_manifest_matching}' in manifest format."
        )

    if missing:
        error("** ERROR: we cannot build some of the requested signatures.")
        error(
            f"** {missing_count} total signatures (for {len(missing)} names) cannot be built."
        )
        if args.ignore_missing:
            error("** (continuing past this error because --ignore-missing was set)")
        else:
            sys.exit(-1)

    notify(
        f"** {total_sigs - skipped_sigs} new signatures to build from {len(to_build)} files;"
    )
    if not to_build:
        notify("** Nothing to build. Exiting!")
        sys.exit(0)

    if skipped_sigs:
        notify(f"** {skipped_sigs} already exist, so skipping those.")
    else:
        notify("** we found no pre-existing signatures that match.")

    ## first, print out a summary of to_build:

    print_results("---")
    print_results("summary of sketches to build:")

    counter = Counter()
    for filename, param_objs in to_build.items():
        for p in param_objs:
            assert len(p.ksizes) == 1
            ksize = p.ksizes[0]
            if not p.dna:
                ksize //= 3

            ski = _SketchInfo(
                ksize=ksize,
                moltype=p.moltype,
                scaled=p.scaled,
                num=p.num_hashes,
                abund=p.track_abundance,
            )
            counter[ski] += 1

    for ski, count in counter.items():
        mh_type = f"num={ski.num}" if ski.num else f"scaled={ski.scaled}"
        mh_abund = ", abund" if ski.abund else ""

        sketch_str = (
            f"{count} sketches with {ski.moltype}, k={ski.ksize}, {mh_type}{mh_abund}"
        )

        print_results(f"   {sketch_str: <50}")

    print_results("---")

    ## now, onward ho - do we build anything, or output stuff, or just exit?

    if args.output_signatures:  # actually compute
        _compute_sigs(
            to_build, args.output_signatures, check_sequence=args.check_sequence
        )

    if args.output_csv_info:  # output info necessary to construct
        _output_csv_info(args.output_csv_info, to_build)

    notify(
        f"** {total_sigs} total requested; output {total_sigs - skipped_sigs}, skipped {skipped_sigs}"
    )


class _signatures_for_compute_factory:
    "Build signatures on demand, based on args input to 'compute'."

    def __init__(self, args):
        self.args = args

    def __call__(self):
        args = self.args
        params = ComputeParameters(
            ksizes=args.ksizes,
            seed=args.seed,
            protein=args.protein,
            dayhoff=args.dayhoff,
            hp=args.hp,
            dna=args.dna,
            num_hashes=args.num_hashes,
            track_abundance=args.track_abundance,
            scaled=args.scaled,
        )
        sig = SourmashSignature.from_params(params)
        return [sig]


def _compute_individual(args, signatures_factory):
    # this is where output signatures will go.
    save_sigs = None

    # track: is this the first file? in cases where we have empty inputs,
    # we don't want to open any outputs.
    first_file_for_output = True

    # if args.output is set, we are aggregating all output to a single file.
    # do not open a new output file for each input.
    open_output_each_time = True
    if args.output:
        open_output_each_time = False

    for filename in args.filenames:
        if open_output_each_time:
            # for each input file, construct output filename
            sigfile = os.path.basename(filename) + ".sig"
            if args.output_dir:
                sigfile = os.path.join(args.output_dir, sigfile)

            # does it already exist? skip if so.
            if os.path.exists(sigfile) and not args.force:
                notify("skipping {} - already done", filename)
                continue  # go on to next file.

            # nope? ok, let's save to it.
            assert not save_sigs
            save_sigs = sourmash_args.SaveSignaturesToLocation(sigfile)

        #
        # calculate signatures!
        #

        # now, set up to iterate over sequences.
        with screed.open(filename) as screed_iter:
            if not screed_iter:
                notify(f"no sequences found in '{filename}'?!")
                continue

            # open output for signatures
            if open_output_each_time:
                save_sigs.open()
            # or... is this the first time to write something to args.output?
            elif first_file_for_output:
                save_sigs = sourmash_args.SaveSignaturesToLocation(args.output)
                save_sigs.open()
                first_file_for_output = False

            # make a new signature for each sequence?
            if args.singleton:
                n_calculated = 0
                for n, record in enumerate(screed_iter):
                    sigs = signatures_factory()
                    try:
                        add_seq(
                            sigs,
                            record.sequence,
                            args.input_is_protein,
                            args.check_sequence,
                        )
                    except ValueError as exc:
                        error(f"ERROR when reading from '{filename}' - ")
                        error(str(exc))
                        sys.exit(-1)

                    n_calculated += len(sigs)
                    set_sig_name(sigs, filename, name=record.name)
                    save_sigs_to_location(sigs, save_sigs)

                notify(
                    "calculated {} signatures for {} sequences in {}",
                    n_calculated,
                    n + 1,
                    filename,
                )

            # nope; make a single sig for the whole file
            else:
                sigs = signatures_factory()

                # consume & calculate signatures
                notify(f"... reading sequences from {filename}")
                name = None
                for n, record in enumerate(screed_iter):
                    if n % 10000 == 0:
                        if n:
                            notify("\r...{} {}", filename, n, end="")
                        elif args.name_from_first:
                            name = record.name

                    try:
                        add_seq(
                            sigs,
                            record.sequence,
                            args.input_is_protein,
                            args.check_sequence,
                        )
                    except ValueError as exc:
                        error(f"ERROR when reading from '{filename}' - ")
                        error(str(exc))
                        sys.exit(-1)

                notify("...{} {} sequences", filename, n, end="")

                set_sig_name(sigs, filename, name)
                save_sigs_to_location(sigs, save_sigs)

                notify(
                    f"calculated {len(sigs)} signatures for {n + 1} sequences in {filename}"
                )

        # if not args.output, close output for every input filename.
        if open_output_each_time:
            save_sigs.close()
            notify(
                f"saved {len(save_sigs)} signature(s) to '{save_sigs.location}'. Note: signature license is CC0."
            )
            save_sigs = None

    # if --output-dir specified, all collected signatures => args.output,
    # and we need to close here.
    if args.output and save_sigs is not None:
        save_sigs.close()
        notify(
            f"saved {len(save_sigs)} signature(s) to '{save_sigs.location}'. Note: signature license is CC0."
        )


def _compute_merged(args, signatures_factory):
    # make a signature for the whole file
    sigs = signatures_factory()

    total_seq = 0
    for filename in args.filenames:
        # consume & calculate signatures
        notify("... reading sequences from {}", filename)

        n = None
        with screed.open(filename) as f:
            for n, record in enumerate(f):
                if n % 10000 == 0 and n:
                    notify("\r... {} {}", filename, n, end="")

                add_seq(
                    sigs, record.sequence, args.input_is_protein, args.check_sequence
                )
        if n is not None:
            notify("... {} {} sequences", filename, n + 1)
            total_seq += n + 1
        else:
            notify(f"no sequences found in '{filename}'?!")

    if total_seq:
        set_sig_name(sigs, filename, name=args.merge)
        notify(
            "calculated 1 signature for {} sequences taken from {} files",
            total_seq,
            len(args.filenames),
        )

        # at end, save!
        save_siglist(sigs, args.output)


def add_seq(sigs, seq, input_is_protein, check_sequence):
    for sig in sigs:
        if input_is_protein:
            sig.add_protein(seq)
        else:
            sig.add_sequence(seq, not check_sequence)


def set_sig_name(sigs, filename, name=None):
    if filename == "-":  # if stdin, set filename to empty.
        filename = ""
    for sig in sigs:
        if name is not None:
            sig._name = name

        sig.filename = filename


def save_siglist(siglist, sigfile_name):
    "Save multiple signatures to a filename."

    # save!
    with sourmash_args.SaveSignaturesToLocation(sigfile_name) as save_sig:
        for ss in siglist:
            save_sig.add(ss)

        notify(f"saved {len(save_sig)} signature(s) to '{save_sig.location}'")


def save_sigs_to_location(siglist, save_sig):
    "Save multiple signatures to an already-open location."
    import sourmash

    for ss in siglist:
        save_sig.add(ss)


class ComputeParameters(RustObject):
    __dealloc_func__ = lib.computeparams_free

    def __init__(
        self,
        *,
        ksizes=(21, 31, 51),
        seed=42,
        protein=False,
        dayhoff=False,
        hp=False,
        dna=True,
        num_hashes=500,
        track_abundance=False,
        scaled=0,
    ):
        self._objptr = lib.computeparams_new()

        self.seed = seed
        self.ksizes = ksizes
        self.protein = protein
        self.dayhoff = dayhoff
        self.hp = hp
        self.dna = dna
        self.num_hashes = num_hashes
        self.track_abundance = track_abundance
        self.scaled = scaled

    @classmethod
    def from_manifest_row(cls, row):
        "convert a CollectionManifest row into a ComputeParameters object"
        is_dna = is_protein = is_dayhoff = is_hp = False
        if row["moltype"] == "DNA":
            is_dna = True
        elif row["moltype"] == "protein":
            is_protein = True
        elif row["moltype"] == "hp":
            is_hp = True
        elif row["moltype"] == "dayhoff":
            is_dayhoff = True
        else:
            assert 0

        if is_dna:
            ksize = row["ksize"]
        else:
            ksize = row["ksize"] * 3

        p = cls(
            ksizes=[ksize],
            seed=DEFAULT_MMHASH_SEED,
            protein=is_protein,
            dayhoff=is_dayhoff,
            hp=is_hp,
            dna=is_dna,
            num_hashes=row["num"],
            track_abundance=row["with_abundance"],
            scaled=row["scaled"],
        )

        return p

    def to_param_str(self):
        "Convert object to equivalent params str."
        pi = []

        if self.dna:
            pi.append("dna")
        elif self.protein:
            pi.append("protein")
        elif self.hp:
            pi.append("hp")
        elif self.dayhoff:
            pi.append("dayhoff")
        else:
            assert 0  # must be one of the previous

        if self.dna:
            kstr = [f"k={k}" for k in self.ksizes]
        else:
            # for protein, divide ksize by three.
            kstr = [f"k={k // 3}" for k in self.ksizes]
        assert kstr
        pi.extend(kstr)

        if self.num_hashes != 0:
            pi.append(f"num={self.num_hashes}")
        elif self.scaled != 0:
            pi.append(f"scaled={self.scaled}")
        else:
            assert 0

        if self.track_abundance:
            pi.append("abund")
        # noabund is default

        if self.seed != DEFAULT_MMHASH_SEED:
            pi.append(f"seed={self.seed}")
        # self.seed

        return ",".join(pi)

    def __repr__(self):
        return f"ComputeParameters(ksizes={self.ksizes}, seed={self.seed}, protein={self.protein}, dayhoff={self.dayhoff}, hp={self.hp}, dna={self.dna}, num_hashes={self.num_hashes}, track_abundance={self.track_abundance}, scaled={self.scaled})"

    def __eq__(self, other):
        return (
            self.ksizes == other.ksizes
            and self.seed == other.seed
            and self.protein == other.protein
            and self.dayhoff == other.dayhoff
            and self.hp == other.hp
            and self.dna == other.dna
            and self.num_hashes == other.num_hashes
            and self.track_abundance == other.track_abundance
            and self.scaled == other.scaled
        )

    @staticmethod
    def from_args(args):
        ptr = lib.computeparams_new()
        ret = ComputeParameters._from_objptr(ptr)

        for arg, value in vars(args).items():
            try:
                getattr(type(ret), arg).fset(ret, value)
            except AttributeError:
                pass

        return ret

    @property
    def seed(self):
        return self._methodcall(lib.computeparams_seed)

    @seed.setter
    def seed(self, v):
        return self._methodcall(lib.computeparams_set_seed, v)

    @property
    def ksizes(self):
        size = ffi.new("uintptr_t *")
        ksizes_ptr = self._methodcall(lib.computeparams_ksizes, size)
        size = size[0]
        ksizes = ffi.unpack(ksizes_ptr, size)
        lib.computeparams_ksizes_free(ksizes_ptr, size)
        return ksizes

    @ksizes.setter
    def ksizes(self, v):
        return self._methodcall(lib.computeparams_set_ksizes, list(v), len(v))

    @property
    def protein(self):
        return self._methodcall(lib.computeparams_protein)

    @protein.setter
    def protein(self, v):
        return self._methodcall(lib.computeparams_set_protein, v)

    @property
    def dayhoff(self):
        return self._methodcall(lib.computeparams_dayhoff)

    @dayhoff.setter
    def dayhoff(self, v):
        return self._methodcall(lib.computeparams_set_dayhoff, v)

    @property
    def hp(self):
        return self._methodcall(lib.computeparams_hp)

    @hp.setter
    def hp(self, v):
        return self._methodcall(lib.computeparams_set_hp, v)

    @property
    def dna(self):
        return self._methodcall(lib.computeparams_dna)

    @dna.setter
    def dna(self, v):
        return self._methodcall(lib.computeparams_set_dna, v)

    @property
    def moltype(self):
        if self.dna:
            moltype = "DNA"
        elif self.protein:
            moltype = "protein"
        elif self.hp:
            moltype = "hp"
        elif self.dayhoff:
            moltype = "dayhoff"
        else:
            assert 0

        return moltype

    @property
    def num_hashes(self):
        return self._methodcall(lib.computeparams_num_hashes)

    @num_hashes.setter
    def num_hashes(self, v):
        return self._methodcall(lib.computeparams_set_num_hashes, v)

    @property
    def track_abundance(self):
        return self._methodcall(lib.computeparams_track_abundance)

    @track_abundance.setter
    def track_abundance(self, v):
        return self._methodcall(lib.computeparams_set_track_abundance, v)

    @property
    def scaled(self):
        return self._methodcall(lib.computeparams_scaled)

    @scaled.setter
    def scaled(self, v):
        return self._methodcall(lib.computeparams_set_scaled, int(v))



"Tests for SqliteIndex, SqliteCollectionManifest, and LCA_SqliteDatabase"

import os
import pytest
import shutil
import sqlite3

import sourmash
from sourmash.exceptions import IndexNotSupported
from sourmash.index.sqlite_index import (
    SqliteIndex,
    load_sqlite_index,
    SqliteCollectionManifest,
    LCA_SqliteDatabase,
)

from sourmash.index import StandaloneManifestIndex
from sourmash import load_one_signature, SourmashSignature
from sourmash.picklist import SignaturePicklist, PickStyle
from sourmash.manifest import CollectionManifest
from sourmash.tax.tax_utils import MultiLineageDB

import sourmash_tst_utils as utils
from sourmash_tst_utils import SourmashCommandFailed
from sourmash import sqlite_utils


def test_sqlite_index_prefetch_empty():
    # check that an exception is raised upon for an empty database
    sig2 = utils.get_test_data("2.fa.sig")
    ss2 = sourmash.load_one_signature(sig2, ksize=31)

    sqlidx = SqliteIndex.create(":memory:")

    # since this is a generator, we need to actually ask for a value to
    # get exception raised.
    g = sqlidx.prefetch(ss2, threshold_bp=0)
    with pytest.raises(ValueError) as e:
        next(g)

    assert "no signatures to search" in str(e.value)


def test_sqlite_index_bad_version(runtmp):
    # create a sqlite database with a bad index version in the
    # sourmash_internal table, see what happens :)

    dbfile = runtmp.output("xyz.sqldb")
    conn = sqlite3.connect(dbfile)
    c = conn.cursor()

    SqliteIndex._create_tables(c)

    # 0.9 doesn't exist/is bad version
    c.execute(
        "UPDATE sourmash_internal SET value=? WHERE key=?", ("0.9", "SqliteIndex")
    )

    conn.commit()

    with pytest.raises(IndexNotSupported):
        sourmash.load_file_as_index(dbfile)


def test_sqlite_index_bad_version_unique(runtmp):
    # try to insert duplicate sqlite index info into sourmash_internal; fail

    dbfile = runtmp.output("xyz.sqldb")
    conn = sqlite3.connect(dbfile)
    c = conn.cursor()

    SqliteIndex._create_tables(c)

    # can't insert duplicate key
    with pytest.raises(sqlite3.IntegrityError):
        c.execute(
            "INSERT INTO sourmash_internal (value, key) VALUES (?, ?)",
            ("1.1", "SqliteIndex"),
        )


def test_index_search_subj_scaled_is_lower():
    # check that subject sketches are appropriately downsampled
    sigfile = utils.get_test_data(
        "scaled100/GCF_000005845.2_ASM584v2_genomic.fna.gz.sig.gz"
    )
    ss = sourmash.load_one_signature(sigfile)

    # double check :)
    assert ss.minhash.scaled == 100

    # build a new query that has a scaled of 1000
    qs = SourmashSignature(ss.minhash.downsample(scaled=1000))

    # create Index to search
    sqlidx = SqliteIndex.create(":memory:")
    sqlidx.insert(ss)

    # search!
    results = list(sqlidx.search(qs, threshold=0))
    assert len(results) == 1
    # original signature (not downsampled) is returned
    assert results[0].signature == ss


def test_sqlite_index_save_load(runtmp):
    sig2 = utils.get_test_data("2.fa.sig")
    sig47 = utils.get_test_data("47.fa.sig")
    sig63 = utils.get_test_data("63.fa.sig")

    ss2 = sourmash.load_one_signature(sig2, ksize=31)
    ss47 = sourmash.load_one_signature(sig47)
    ss63 = sourmash.load_one_signature(sig63)

    filename = runtmp.output("foo")
    sqlidx = SqliteIndex.create(filename)
    sqlidx.insert(ss2)
    sqlidx.insert(ss47)
    sqlidx.insert(ss63)

    sqlidx.close()

    sqlidx2 = SqliteIndex.load(filename)

    # now, search for sig2
    sr = sqlidx2.search(ss2, threshold=1.0)
    print([s[1].name for s in sr])
    assert len(sr) == 1
    assert sr[0][1] == ss2


def test_sqlite_index_multik_select():
    # this loads three ksizes, 21/31/51
    sig2 = utils.get_test_data("2.fa.sig")
    siglist = sourmash.load_file_as_signatures(sig2)

    sqlidx = SqliteIndex.create(":memory:")
    for ss in siglist:
        sqlidx.insert(ss)

    # select most specifically
    sqlidx2 = sqlidx.select(ksize=31, moltype="DNA")
    assert len(sqlidx2) == 1

    # all are DNA:
    sqlidx2 = sqlidx.select(moltype="DNA")
    assert len(sqlidx2) == 3


def test_sqlite_index_num_select():
    # this will fail on 'num' select, which is not allowed
    sqlidx = SqliteIndex.create(":memory:")
    with pytest.raises(ValueError):
        sqlidx.select(num=100)


def test_sqlite_index_abund_select():
    # this will fail on 'track_abundance' select, which is not allowed
    sqlidx = SqliteIndex.create(":memory:")
    with pytest.raises(ValueError):
        sqlidx.select(track_abundance=True)


def test_sqlite_index_insert_num_fail():
    # cannot insert 'num' signatures
    sqlidx = SqliteIndex.create(":memory:")

    sig47 = utils.get_test_data("num/47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)
    assert ss47.minhash.num != 0

    with pytest.raises(ValueError) as exc:
        sqlidx.insert(ss47)

    assert "cannot store 'num' signatures in SqliteIndex" in str(exc)


def test_sqlite_index_insert_abund_fail():
    # cannot insert 'num' signatures
    sqlidx = SqliteIndex.create(":memory:")

    sig47 = utils.get_test_data("track_abund/47.fa.sig")
    ss47 = sourmash.load_one_signature(sig47, ksize=31)

    with pytest.raises(ValueError) as exc:
        sqlidx.insert(ss47)

    assert "cannot store signatures with abundance in SqliteIndex" in str(exc)


def test_sqlite_index_moltype_multi_fail():
    # check that we cannot store sigs with multiple scaled values.

    # this loads multiple ksizes (19, 31) and moltypes (DNA, protein, hp, etc)
    filename = utils.get_test_data("prot/all.zip")
    siglist = sourmash.load_file_as_signatures(filename)
    siglist = list(siglist)

    sqlidx = SqliteIndex.create(":memory:")

    sqlidx.insert(siglist[0])
    assert sqlidx.scaled == 100

    with pytest.raises(ValueError) as exc:
        for ss in siglist:
            sqlidx.insert(ss)

    assert "this database can only store scaled values=100" in str(exc)


def test_sqlite_index_picklist_select():
    # test select with a picklist

    # this loads three ksizes, 21/31/51
    sig2 = utils.get_test_data("2.fa.sig")
    siglist = sourmash.load_file_as_signatures(sig2)

    sqlidx = SqliteIndex.create(":memory:")
    for ss in siglist:
        sqlidx.insert(ss)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8")
    picklist.init(["f3a90d4e"])

    # select on picklist
    sqlidx2 = sqlidx.select(picklist=picklist)
    assert len(sqlidx2) == 1
    ss = list(sqlidx2.signatures())[0]
    assert ss.minhash.ksize == 31
    assert ss.md5sum().startswith("f3a90d4e55")


def test_sqlite_index_picklist_select_exclude():
    # test select with a picklist, but exclude

    # this loads three ksizes, 21/31/51
    sig2 = utils.get_test_data("2.fa.sig")
    siglist = sourmash.load_file_as_signatures(sig2)

    sqlidx = SqliteIndex.create(":memory:")
    for ss in siglist:
        sqlidx.insert(ss)

    # construct a picklist...
    picklist = SignaturePicklist("md5prefix8", pickstyle=PickStyle.EXCLUDE)
    picklist.init(["f3a90d4e"])

    # select on picklist
    sqlidx2 = sqlidx.select(picklist=picklist)
    assert len(sqlidx2) == 2
    md5s = set()
    ksizes = set()
    for ss in list(sqlidx2.signatures()):
        md5s.add(ss.md5sum())
        ksizes.add(ss.minhash.ksize)
    assert md5s == set(
        ["f372e47893edd349e5956f8b0d8dcbf7", "43f3b48e59443092850964d355a20ac0"]
    )
    assert ksizes == set([21, 51])


def test_sqlite_jaccard_ordering():
    # this tests a tricky situation where for three sketches A, B, C,
    # |A intersect B| is greater than |A intersect C|
    # _but_
    # |A jaccard B| is less than |A intersect B|
    a = sourmash.MinHash(ksize=31, n=0, scaled=2)
    b = a.copy_and_clear()
    c = a.copy_and_clear()

    a.add_many([1, 2, 3, 4])
    b.add_many([1, 2, 3] + list(range(10, 30)))
    c.add_many([1, 5])

    def _intersect(x, y):
        return x.intersection_and_union_size(y)[0]

    print("a intersect b:", _intersect(a, b))
    print("a intersect c:", _intersect(a, c))
    print("a jaccard b:", a.jaccard(b))
    print("a jaccard c:", a.jaccard(c))
    assert _intersect(a, b) > _intersect(a, c)
    assert a.jaccard(b) < a.jaccard(c)

    # thresholds to use:
    assert a.jaccard(b) < 0.15
    assert a.jaccard(c) > 0.15

    # now - make signatures, try out :)
    ss_a = sourmash.SourmashSignature(a, name="A")
    ss_b = sourmash.SourmashSignature(b, name="B")
    ss_c = sourmash.SourmashSignature(c, name="C")

    sqlidx = SqliteIndex.create(":memory:")
    sqlidx.insert(ss_a)
    sqlidx.insert(ss_b)
    sqlidx.insert(ss_c)

    sr = sqlidx.search(ss_a, threshold=0.15)
    print(sr)
    assert len(sr) == 2
    assert sr[0].signature == ss_a
    assert sr[0].score == 1.0
    assert sr[1].signature == ss_c
    assert sr[1].score == 0.2


def test_sqlite_index_scaled1():
    # check on scaled=1 storage.
    sqlidx = SqliteIndex.create(":memory:")

    mh1 = sourmash.MinHash(0, 31, scaled=1)
    mh1.add_hash(2**64 - 1)
    mh1.add_hash(2**64 - 2)
    mh1.add_hash(2**64 - 3)
    ss1 = sourmash.SourmashSignature(mh1, name="ss 1")

    mh2 = sourmash.MinHash(0, 31, scaled=1)
    mh2.add_hash(2**64 - 1)
    mh2.add_hash(2**64 - 2)
    mh2.add_hash(2**64 - 3)
    mh2.add_hash(0)
    mh2.add_hash(1)
    mh2.add_hash(2)
    ss2 = sourmash.SourmashSignature(mh2, name="ss 2")

    sqlidx.insert(ss1)
    sqlidx.insert(ss2)

    # check jaccard search
    results = list(sqlidx.search(ss1, threshold=0))
    print(results)
    assert len(results) == 2
    assert results[0].signature == ss1
    assert results[0].score == 1.0
    assert results[1].signature == ss2
    assert results[1].score == 0.5

    results = list(sqlidx.search(ss1, threshold=0, do_containment=True))
    print(results)
    assert results[0].signature == ss1
    assert results[0].score == 1.0
    assert results[1].signature == ss2
    assert results[1].score == 1.0

    # minhashes retrieved successfully?
    assert len(results[0].signature.minhash) == 3
    assert len(results[1].signature.minhash) == 6


def test_sqlite_index_load_existing():
    # try loading an existing sqlite index
    filename = utils.get_test_data("sqlite/index.sqldb")
    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, SqliteIndex)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2


def test_sqlite_index_create_load_existing(runtmp):
    # try creating then loading an existing sqlite index; create from CLI
    filename = runtmp.output("idx.sqldb")
    sig1 = utils.get_test_data("47.fa.sig")
    sig2 = utils.get_test_data("63.fa.sig")

    runtmp.sourmash("sig", "cat", sig1, sig2, "-o", filename)

    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, SqliteIndex)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2


def test_sqlite_index_create_load_insert_existing(runtmp):
    # try creating, loading, inserting into an existing sqlite index
    filename = runtmp.output("idx.sqldb")
    sig1 = utils.get_test_data("47.fa.sig")
    sig2 = utils.get_test_data("63.fa.sig")
    sig3 = utils.get_test_data("2.fa.sig")

    runtmp.sourmash("sig", "cat", sig1, sig2, "-o", filename)

    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, SqliteIndex)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2

    ss3 = sourmash.load_one_signature(sig3, ksize=31)
    sqlidx.insert(ss3)
    sqlidx.commit()

    runtmp.sourmash("sig", "describe", filename)
    print(runtmp.last_result.out)
    assert "md5: f3a90d4e5528864a5bcc8434b0d0c3b1" in runtmp.last_result.out


def test_sqlite_index_create_load_insert_existing_cli(runtmp):
    # try creating, loading, inserting into an existing sqlite index from cli
    # (aka "append" to existing database)
    filename = runtmp.output("idx.sqldb")
    sig1 = utils.get_test_data("47.fa.sig")
    sig2 = utils.get_test_data("63.fa.sig")
    sig3 = utils.get_test_data("2.fa.sig")

    runtmp.sourmash("sig", "cat", sig1, sig2, "-o", filename)

    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, SqliteIndex)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2

    # add a third
    runtmp.sourmash("sig", "cat", sig3, "-o", filename, "-k", "31")

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 3


def test_sqlite_manifest_bad_version(runtmp):
    # create a sqlite database with a bad manifest version in the
    # sourmash_internal table, see what happens :)

    dbfile = runtmp.output("xyz.sqlmf")
    conn = sqlite3.connect(dbfile)
    c = conn.cursor()

    SqliteCollectionManifest._create_tables(c)

    # 0.9 doesn't exist/bad version
    c.execute(
        "UPDATE sourmash_internal SET value=? WHERE key=?", ("0.9", "SqliteManifest")
    )

    conn.commit()

    with pytest.raises(IndexNotSupported):
        CollectionManifest.load_from_filename(dbfile)


def test_sqlite_manifest_bad_version_unique(runtmp):
    # try to insert duplicate sqlite manifest info into sourmash_internal; fail

    dbfile = runtmp.output("xyz.sqldb")
    conn = sqlite3.connect(dbfile)
    c = conn.cursor()

    SqliteCollectionManifest._create_tables(c)

    # can't insert duplicate key
    with pytest.raises(sqlite3.IntegrityError):
        c.execute(
            "INSERT INTO sourmash_internal (value, key) VALUES (?, ?)",
            ("1.1", "SqliteManifest"),
        )


def test_sqlite_manifest_basic():
    # test some features of the SQLite-based manifest.
    sig2 = load_one_signature(utils.get_test_data("2.fa.sig"), ksize=31)
    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"), ksize=31)
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"), ksize=31)

    sqlidx = SqliteIndex.create(":memory:")

    # empty manifest tests
    manifest = sqlidx.manifest
    assert not manifest
    assert len(manifest) == 0

    sqlidx.insert(sig47)
    sqlidx.insert(sig63)

    # ok, more full manifest tests!
    assert manifest
    assert len(manifest) == 2

    assert sig47 in manifest
    assert sig2 not in manifest

    # check that we can get a "standard" manifest out
    standard_mf = CollectionManifest.load_from_manifest(sqlidx.manifest)
    assert len(standard_mf) == 2

    picklist = manifest.to_picklist()
    assert sig47 in picklist
    assert sig2 not in picklist


def test_sqlite_manifest_round_trip():
    # check that we can go from regular mf -> sqlite mf -> regular again.
    sig2 = load_one_signature(utils.get_test_data("2.fa.sig"), ksize=31)
    sig47 = load_one_signature(utils.get_test_data("47.fa.sig"), ksize=31)
    sig63 = load_one_signature(utils.get_test_data("63.fa.sig"), ksize=31)

    rows = []
    rows.append(
        CollectionManifest.make_manifest_row(sig47, None, include_signature=False)
    )
    rows.append(
        CollectionManifest.make_manifest_row(sig63, None, include_signature=False)
    )
    nosql_mf = CollectionManifest(rows)

    sqlite_mf = SqliteCollectionManifest.load_from_manifest(nosql_mf)

    # test roundtrip
    round_mf = CollectionManifest.load_from_manifest(sqlite_mf)

    assert len(round_mf) == 2
    print(round_mf.rows, nosql_mf.rows)
    assert round_mf == nosql_mf

    for mf in (nosql_mf, sqlite_mf, round_mf):
        picklist = mf.to_picklist()
        assert sig47 in picklist
        assert sig2 not in picklist


def test_sqlite_manifest_create(runtmp):
    # test creation and summarization of a manifest of prot.zip
    zipfile = utils.get_test_data("prot/all.zip")

    # create manifest
    runtmp.sourmash("sig", "manifest", "-F", "sql", zipfile, "-o", "mf.sqlmf")

    sqlmf = runtmp.output("mf.sqlmf")
    assert os.path.exists(sqlmf)

    # verify it's loadable as the right type
    idx = load_sqlite_index(sqlmf)
    assert isinstance(idx, StandaloneManifestIndex)

    # summarize
    runtmp.sourmash("sig", "fileinfo", "mf.sqlmf")

    out = runtmp.last_result.out
    print(out)

    assert "2 sketches with dayhoff, k=19, scaled=100          7945 total hashes" in out
    assert "2 sketches with hp, k=19, scaled=100               5184 total hashes" in out
    assert "2 sketches with protein, k=19, scaled=100          8214 total hashes" in out
    assert "1 sketches with DNA, k=31, scaled=1000             5238 total hashes" in out

    assert "path filetype: StandaloneManifestIndex" in out
    assert "location: mf.sqlmf" in out
    assert "is database? yes" in out
    assert "has manifest? yes" in out
    assert "num signatures: 7" in out


def test_sqlite_manifest_create_noload_sigs(runtmp):
    # sigs should not be loadable from manifest this way...
    zipfile = utils.get_test_data("prot/all.zip")

    # create manifest
    runtmp.sourmash("sig", "manifest", "-F", "sql", zipfile, "-o", "mf.sqlmf")

    # 'describe' should not be able to load the sqlmf b/c prefix is wrong
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash("sig", "describe", "mf.sqlmf")


def test_sqlite_manifest_create_yesload_sigs(runtmp):
    # should be able to load after copying files
    zipfile = utils.get_test_data("prot/all.zip")
    shutil.copytree(utils.get_test_data("prot"), runtmp.output("prot"))

    # create manifest
    runtmp.sourmash("sig", "manifest", "-F", "sql", zipfile, "-o", "prot/mf.sqlmf")

    # 'describe' should now be able to load the sqlmf, which is cool
    runtmp.sourmash("sig", "describe", "prot/mf.sqlmf")
    print(runtmp.last_result.out)


def test_sqlite_manifest_num(runtmp):
    # should be able to produce sql manifests with 'num' sketches in them
    numsig = utils.get_test_data("num/47.fa.sig")

    # create mf
    runtmp.sourmash("sig", "manifest", "-F", "sql", numsig, "-o", "mf.sqlmf")

    # do summarize:
    runtmp.sourmash("sig", "summarize", "mf.sqlmf")
    out = runtmp.last_result.out

    print(out)

    assert "1 sketches with DNA, k=21, num=500                 500 total hashes" in out
    assert "1 sketches with DNA, k=31, num=500                 500 total hashes" in out
    assert "1 sketches with DNA, k=51, num=500                 500 total hashes" in out


def test_sqlite_manifest_num_select(runtmp):
    # should be able to _select_ sql manifests with 'num' sketches in them
    numsig = utils.get_test_data("num/47.fa.sig")

    # create mf
    runtmp.sourmash("sig", "manifest", "-F", "sql", numsig, "-o", "mf.sqlmf")

    # load as index
    idx = sourmash.load_file_as_index(runtmp.output("mf.sqlmf"))

    # select
    print(list(idx.manifest.rows))
    idx = idx.select(num=500)
    print(list(idx.manifest.rows))
    assert len(idx) == 3


def test_sqlite_manifest_locations(runtmp):
    # check what locations returns... may return too many, that's ok.
    prot = utils.get_test_data("prot")

    runtmp.sourmash("sig", "manifest", "-F", "sql", prot, "-o", "mf.sqlmf")

    # load as index
    idx = sourmash.load_file_as_index(runtmp.output("mf.sqlmf"))

    picklist = SignaturePicklist("identprefix")
    picklist.pickset = set(["GCA_001593925"])
    idx = idx.select(picklist=picklist)

    sql_locations = set(idx.manifest.locations())
    row_locations = set(row["internal_location"] for row in idx.manifest.rows)

    assert sql_locations.issuperset(row_locations)

    assert "dna-sig.sig.gz" in sql_locations  # this is unnecessary...
    assert "dna-sig.sig.gz" not in row_locations  # ...this is correct :)


def test_sqlite_manifest_create_insert(runtmp):
    # try out creating a sqlite manifest and then running cli on it

    mfname = runtmp.output("some.sqlmf")
    mf = SqliteCollectionManifest.create(mfname)

    sigfile = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sigfile)

    mf._insert_row(mf.conn.cursor(), mf.make_manifest_row(ss, "some.sig"))
    mf.conn.commit()

    # copy sig in since we want it to resolve...
    shutil.copyfile(sigfile, runtmp.output("some.sig"))

    # 'describe' should work here, to resolve actual sigs.
    runtmp.sourmash("sig", "describe", mfname)
    print(runtmp.last_result.out)
    assert "md5: 09a08691ce52952152f0e866a59f6261" in runtmp.last_result.out


def test_sqlite_manifest_create_insert_2(runtmp):
    # try out creating a sqlite manifest from cli and then _insert_row into it

    # copy sig in since we want it to resolve...
    sigfile = utils.get_test_data("47.fa.sig")
    shutil.copyfile(sigfile, runtmp.output("some.sig"))

    runtmp.sourmash("sig", "manifest", "some.sig", "-F", "sql", "-o", "some.sqlmf")
    mfname = runtmp.output("some.sqlmf")

    mf = CollectionManifest.load_from_filename(mfname)
    ss = sourmash.load_one_signature(runtmp.output("some.sig"))
    mf._insert_row(mf.conn.cursor(), mf.make_manifest_row(ss, "some.sig"))
    mf.conn.commit()

    # 'describe' should work here, to resolve actual sigs.
    runtmp.sourmash("sig", "describe", mfname)
    print(runtmp.last_result.out)
    assert "md5: 09a08691ce52952152f0e866a59f6261" in runtmp.last_result.out


def test_sqlite_manifest_existing(runtmp):
    # try out an existing sqlite manifest

    prefix = runtmp.output("protdir")
    mf = runtmp.output("protdir/prot.sqlmf")
    shutil.copytree(utils.get_test_data("prot"), prefix)
    shutil.copyfile(utils.get_test_data("sqlite/prot.sqlmf"), mf)

    runtmp.sourmash("sig", "describe", mf)
    print(runtmp.last_result.out)


def test_sqlite_manifest_existing_insert(runtmp):
    # try out an existing sqlite manifest - insert into it

    prefix = runtmp.output("protdir")
    shutil.copytree(utils.get_test_data("prot"), prefix)

    mfname = runtmp.output("protdir/prot.sqlmf")
    shutil.copyfile(utils.get_test_data("sqlite/prot.sqlmf"), mfname)
    mf = CollectionManifest.load_from_filename(mfname)
    assert isinstance(mf, SqliteCollectionManifest)

    sigfile = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sigfile)

    mf._insert_row(mf.conn.cursor(), mf.make_manifest_row(ss, "some.sig"))
    mf.conn.commit()

    # copy sig in since we want it to resolve...
    shutil.copyfile(sigfile, runtmp.output("protdir/some.sig"))

    # 'describe' should work here.
    runtmp.sourmash("sig", "describe", mfname)
    print(runtmp.last_result.out)


def test_sqlite_manifest_existing_mf_only(runtmp):
    # try out an existing sqlite manifest, but without underlying files -> fail

    mf = runtmp.output("prot.sqlmf")
    shutil.copyfile(utils.get_test_data("sqlite/prot.sqlmf"), mf)

    # 'fileinfo' should work...
    runtmp.sourmash("sig", "fileinfo", mf)
    print(runtmp.last_result.out)
    assert "num signatures: 7" in runtmp.last_result.out

    # ...but 'describe' should fail, since it needs actual sigs.
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash("sig", "describe", mf)

    print(runtmp.last_result.err)
    assert "ERROR: Error while reading signatures from" in runtmp.last_result.err


def test_sqlite_manifest_existing_mfonly_insert(runtmp):
    # try out an existing sqlite manifest - insert into it, but fail describe

    mfname = runtmp.output("prot.sqlmf")
    shutil.copyfile(utils.get_test_data("sqlite/prot.sqlmf"), mfname)
    mf = CollectionManifest.load_from_filename(mfname)
    assert isinstance(mf, SqliteCollectionManifest)

    sigfile = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sigfile)

    mf._insert_row(mf.conn.cursor(), mf.make_manifest_row(ss, sigfile))
    mf.conn.commit()

    # 'fileinfo' should work...
    runtmp.sourmash("sig", "fileinfo", mfname)
    print(runtmp.last_result.out)
    assert "num signatures: 8" in runtmp.last_result.out

    # ...but 'describe' should fail, since it needs actual sigs.
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash("sig", "describe", mfname)


def test_sqlite_manifest_load_existing_index():
    # try loading an existing sqlite index as a manifest
    filename = utils.get_test_data("sqlite/index.sqldb")
    mf = CollectionManifest.load_from_filename(filename)
    assert isinstance(mf, SqliteCollectionManifest)

    assert len(mf) == 2


def test_sqlite_manifest_load_existing_index_insert_fail():
    # try loading an existing sqlite index as a manifest; insert should fail
    filename = utils.get_test_data("sqlite/index.sqldb")
    mf = CollectionManifest.load_from_filename(filename)
    assert isinstance(mf, SqliteCollectionManifest)

    assert len(mf) == 2

    # try insert - should fail
    sigfile = utils.get_test_data("47.fa.sig")
    ss = sourmash.load_one_signature(sigfile)

    with pytest.raises(Exception) as exc:
        mf._insert_row(mf.conn.cursor(), mf.make_manifest_row(ss, sigfile))

    assert "must use SqliteIndex.insert to add to this manifest" in str(exc)


def test_sqlite_manifest_create_load_empty(runtmp):
    # try creating an empty manifest, then loading
    mfname = runtmp.output("some.sqlmf")
    mf = SqliteCollectionManifest.create(mfname)
    mf.close()

    mf2 = load_sqlite_index(mfname)
    assert len(mf2) == 0


def test_sqlite_lca_db_load_existing():
    # try loading an existing sqlite index
    filename = utils.get_test_data("sqlite/lca.sqldb")
    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, LCA_SqliteDatabase)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2


def test_sqlite_lca_db_select():
    # try loading an existing sqlite index
    filename = utils.get_test_data("sqlite/lca.sqldb")
    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, LCA_SqliteDatabase)

    sqlidx2 = sqlidx.select(ksize=31)
    list(sqlidx2.hashvals)  # only on LCA_SqliteDatabase
    assert isinstance(sqlidx2, LCA_SqliteDatabase)


def test_sqlite_lca_db_create_load_existing(runtmp):
    # try creating (from CLI) then loading (from API) an LCA db
    filename = runtmp.output("lca.sqldb")
    sig1 = utils.get_test_data("lca/TARA_ASE_MAG_00031.sig")
    sig2 = utils.get_test_data("lca/TARA_PSW_MAG_00136.sig")

    runtmp.sourmash("sig", "flatten", sig1, sig2, "-o", filename, "-k", "31")

    # load tax
    tax_csv = utils.get_test_data("sqlite/delmont-6.csv")
    runtmp.sourmash("tax", "prepare", "-t", tax_csv, "-o", filename, "-F", "sql")

    sqlidx = sourmash.load_file_as_index(filename)
    assert isinstance(sqlidx, LCA_SqliteDatabase)

    siglist = list(sqlidx.signatures())
    assert len(siglist) == 2


def test_sqlite_lca_db_load_empty(runtmp):
    # try creating then loading an _empty_ LCA_SqliteDatabase

    dbname = runtmp.output("empty.sqldb")

    # create empty SqliteIndex...
    runtmp.sourmash("sig", "cat", "-o", dbname)
    assert os.path.exists(dbname)

    # ...and create empty sourmash_taxonomy tables in there...
    empty_tax = utils.get_test_data("scaled/empty-lineage.csv")
    runtmp.sourmash("tax", "prepare", "-F", "sql", "-t", empty_tax, "-o", dbname)

    runtmp.sourmash("sig", "describe", dbname)
    assert "loaded 0 signatures" in runtmp.last_result.err


def test_sqlite_lca_db_create_readonly(runtmp):
    # try running 'prepare' on a read-only sqlite db, check error message.

    dbname = runtmp.output("empty.sqldb")

    # create empty SqliteIndex...
    runtmp.sourmash("sig", "cat", "-o", dbname)
    assert os.path.exists(dbname)

    # make it read only...
    from stat import S_IREAD, S_IRGRP, S_IROTH

    os.chmod(dbname, S_IREAD | S_IRGRP | S_IROTH)

    # ...and try creating empty sourmash_taxonomy tables in there...
    empty_tax = utils.get_test_data("scaled/empty-lineage.csv")

    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash("tax", "prepare", "-F", "sql", "-t", empty_tax, "-o", dbname)

    err = runtmp.last_result.err
    print(err)

    assert "taxonomy table already exists in" not in err
    assert "attempt to write a readonly database" in err


def test_sqlite_lca_db_try_load_sqlite_index():
    # try loading a SqliteIndex with no tax tables from .load classmethod
    dbname = utils.get_test_data("sqlite/index.sqldb")

    with pytest.raises(ValueError) as exc:
        LCA_SqliteDatabase.load(dbname)

    assert "not a taxonomy database" in str(exc)


def test_sqlite_lca_db_supply_lineage_db():
    # try creating an LCA_SqliteDatabase object with a separate lineage DB.
    dbname = utils.get_test_data("sqlite/index.sqldb")

    tax_csv = utils.get_test_data("sqlite/shewanella-lineage.csv")
    lineage_db = MultiLineageDB.load([tax_csv])

    db = LCA_SqliteDatabase(dbname, lineage_db=lineage_db)

    hashval = next(iter(db.hashvals))
    lineages = db.get_lineage_assignments(hashval)
    print(lineages)
    assert lineages[0][0].rank == "superkingdom"
    assert lineages[0][0].name == "d__Bacteria"
    assert lineages[0][-1].rank == "species"
    assert lineages[0][-1].name == "s__Shewanella baltica"
    assert lineages[1][0].rank == "superkingdom"
    assert lineages[1][0].name == "d__Bacteria"
    assert lineages[0][-1].rank == "species"
    assert lineages[0][-1].name == "s__Shewanella baltica"


def test_bad_sqlite_internal_version():
    # check get_sourmash_internal
    dbname = utils.get_test_data("sqlite/index.sqldb")

    conn = sqlite_utils.open_sqlite_db(dbname)
    c = conn.cursor()
    with pytest.raises(Exception):
        sqlite_utils.add_sourmash_internal(c, "SqliteIndex", "0.9")



"""
Tests for functions in lca submodule.
"""

import pytest

from sourmash.lca import lca_utils
from sourmash.lca.lca_utils import (
    LineagePair,
    build_tree,
    find_lca,
    taxlist,
    count_lca_for_assignments,
    zip_lineage,
    display_lineage,
    make_lineage,
    is_lineage_match,
    pop_to_rank,
)


class FakeLCA_Database:
    def __init__(self):
        self._assignments = {}

    def _set_lineage_assignment(self, hashval, assignment):
        self._assignments[hashval] = assignment

    def get_lineage_assignments(self, hashval):
        if hashval in self._assignments:
            return self._assignments[hashval]
        else:
            return None


def test_taxlist_1():
    assert list(taxlist()) == [
        "superkingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
        "strain",
    ]


def test_taxlist_2():
    assert list(taxlist(include_strain=False)) == [
        "superkingdom",
        "phylum",
        "class",
        "order",
        "family",
        "genus",
        "species",
    ]


def test_zip_lineage_1():
    x = [LineagePair("superkingdom", "a"), LineagePair("phylum", "b")]
    assert zip_lineage(x) == ["a", "b", "", "", "", "", "", ""]


def test_zip_lineage_2():
    x = [LineagePair("superkingdom", "a"), LineagePair("phylum", "b")]
    assert zip_lineage(x, truncate_empty=True) == ["a", "b"]


def test_zip_lineage_3():
    x = [
        LineagePair("superkingdom", "a"),
        LineagePair(None, ""),
        LineagePair("class", "c"),
    ]
    assert zip_lineage(x) == ["a", "", "c", "", "", "", "", ""]


def test_zip_lineage_3_truncate():
    x = [
        LineagePair("superkingdom", "a"),
        LineagePair(None, ""),
        LineagePair("class", "c"),
    ]
    assert zip_lineage(x, truncate_empty=True) == ["a", "", "c"]


def test_zip_lineage_4():
    x = [LineagePair("superkingdom", "a"), LineagePair("class", "c")]
    with pytest.raises(ValueError) as e:
        zip_lineage(x)

    assert "incomplete lineage at phylum - is class instead" in str(e.value)


def test_display_lineage_1():
    x = [LineagePair("superkingdom", "a"), LineagePair("phylum", "b")]
    assert display_lineage(x) == "a;b", display_lineage(x)


def test_display_lineage_2():
    x = [
        LineagePair("superkingdom", "a"),
        LineagePair(None, ""),
        LineagePair("class", "c"),
    ]
    assert display_lineage(x) == "a;;c", display_lineage(x)


def test_build_tree():
    tree = build_tree([[LineagePair("rank1", "name1"), LineagePair("rank2", "name2")]])
    assert tree == {LineagePair("rank1", "name1"): {LineagePair("rank2", "name2"): {}}}


def test_build_tree_2():
    tree = build_tree(
        [
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2a")],
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2b")],
        ]
    )

    assert tree == {
        LineagePair("rank1", "name1"): {
            LineagePair("rank2", "name2a"): {},
            LineagePair("rank2", "name2b"): {},
        }
    }


def test_build_tree_3():  # empty 'rank2' name
    tree = build_tree([[LineagePair("rank1", "name1"), LineagePair("rank2", "")]])
    assert tree == {LineagePair("rank1", "name1"): {}}


def test_build_tree_4():
    tree = build_tree(
        [
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2a")],
        ]
    )

    tree = build_tree(
        [
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2b")],
        ],
        tree,
    )

    assert tree == {
        LineagePair("rank1", "name1"): {
            LineagePair("rank2", "name2a"): {},
            LineagePair("rank2", "name2b"): {},
        }
    }


def test_build_tree_5():
    with pytest.raises(ValueError):
        build_tree([])


def test_find_lca():
    tree = build_tree([[LineagePair("rank1", "name1"), LineagePair("rank2", "name2")]])
    lca = find_lca(tree)

    assert lca == (
        (
            LineagePair("rank1", "name1"),
            LineagePair("rank2", "name2"),
        ),
        0,
    )


def test_find_lca_2():
    tree = build_tree(
        [
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2a")],
            [LineagePair("rank1", "name1"), LineagePair("rank2", "name2b")],
        ]
    )
    lca = find_lca(tree)

    assert lca == ((LineagePair("rank1", "name1"),), 2)


def test_find_lca_3():
    lin1 = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b")

    tree = build_tree([lin1, lin2])
    lca, reason = find_lca(tree)
    assert lca == lin1  # find most specific leaf node


def test_gather_assignments_1():
    # test basic mechanics of gather_assignments function
    hashval = 12345678
    lin = lca_utils.make_lineage("a;b;c")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin]))

    assignments = lca_utils.gather_assignments([hashval], [db])
    print(assignments)

    assert assignments[hashval] == set([lin])


def test_gather_assignments_2():
    # test basic mechanics of gather_assignments function with two lineages
    hashval = 12345678
    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))

    assignments = lca_utils.gather_assignments([hashval], [db])
    print(assignments)

    assert assignments[hashval] == set([lin, lin2])


def test_gather_assignments_3():
    # test basic mechanics of gather_assignments function with two lineages
    # and two hashvals
    hashval = 12345678
    hashval2 = 87654321
    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))
    db._set_lineage_assignment(hashval2, set([lin]))

    assignments = lca_utils.gather_assignments([hashval, hashval2], [db])
    print(assignments)

    assert assignments[hashval] == set([lin, lin2])
    assert assignments[hashval2] == set([lin])


def test_count_lca_for_assignments_1():
    # test basic mechanics of gather_assignments function
    hashval = 12345678
    lin = lca_utils.make_lineage("a;b;c")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin]))

    assignments = lca_utils.gather_assignments([hashval], [db])
    counts = count_lca_for_assignments(assignments)
    print(counts)

    assert len(counts) == 1
    assert counts[lin] == 1


def test_count_lca_for_assignments_2():
    # test basic mechanics of gather_assignments function with two lineages
    hashval = 12345678
    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))

    assignments = lca_utils.gather_assignments([hashval], [db])
    counts = count_lca_for_assignments(assignments)
    print(counts)

    assert counts[lin] == 0
    assert counts[lin2] == 0

    assert len(counts) == 1
    lca_lin = lca_utils.make_lineage("a;b")
    assert counts[lca_lin] == 1


def test_count_lca_for_assignments_3():
    # test basic mechanics of gather_assignments function with two lineages
    # and two hashvals
    hashval = 12345678
    hashval2 = 87654321
    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))
    db._set_lineage_assignment(hashval2, set([lin]))

    assignments = lca_utils.gather_assignments([hashval, hashval2], [db])
    counts = count_lca_for_assignments(assignments)
    print(counts)

    assert len(counts) == 2
    assert counts[lin] == 1
    assert counts[lin2] == 0

    lca_lin = lca_utils.make_lineage("a;b")
    assert counts[lca_lin] == 1


def test_count_lca_for_assignments_abund_1():
    # test basic mechanics of gather_assignments function
    hashval = 12345678
    hashval_counts = {}
    hashval_counts[hashval] = 3

    lin = lca_utils.make_lineage("a;b;c")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin]))

    assignments = lca_utils.gather_assignments(hashval_counts.keys(), [db])
    counts = count_lca_for_assignments(assignments, hashval_counts)
    print(counts)

    assert len(counts) == 1
    assert counts[lin] == 3


def test_count_lca_for_assignments_abund_2():
    # test basic mechanics of gather_assignments function with two lineages
    hashval = 12345678
    hashval_counts = {}
    hashval_counts[hashval] = 3

    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))

    assignments = lca_utils.gather_assignments(hashval_counts, [db])
    counts = count_lca_for_assignments(assignments, hashval_counts)
    print(counts)

    assert counts[lin] == 0
    assert counts[lin2] == 0

    assert len(counts) == 1
    lca_lin = lca_utils.make_lineage("a;b")
    assert counts[lca_lin] == 3  # yes!


def test_count_lca_for_assignments_abund_3():
    # test basic mechanics of gather_assignments function with two lineages
    # and two hashvals
    hashval = 12345678
    hashval2 = 87654321
    hashval_counts = {}
    hashval_counts[hashval] = 2
    hashval_counts[hashval2] = 5

    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))
    db._set_lineage_assignment(hashval2, set([lin]))

    assignments = lca_utils.gather_assignments(hashval_counts, [db])
    counts = count_lca_for_assignments(assignments, hashval_counts)
    print(counts)

    assert len(counts) == 2
    assert counts[lin] == 5  # makes sense
    assert counts[lin2] == 0  # makes sense

    lca_lin = lca_utils.make_lineage("a;b")
    assert counts[lca_lin] == 2  # yes!


def test_count_lca_for_assignments_abund_4():
    # test basic mechanics of gather_assignments function with three lineages
    # and three hashvals
    hashval = 12345678
    hashval2 = 87654321
    hashval3 = 34567891
    hashval_counts = {}
    hashval_counts[hashval] = 2
    hashval_counts[hashval2] = 5
    hashval_counts[hashval3] = 3

    lin = lca_utils.make_lineage("a;b;c")
    lin2 = lca_utils.make_lineage("a;b;d")
    lin3 = lca_utils.make_lineage("a;b;d;e")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))  # lca: a;b
    db._set_lineage_assignment(hashval2, set([lin]))  # lca: a;b;c
    db._set_lineage_assignment(hashval3, set([lin2, lin3]))  # a;b;d;e

    assignments = lca_utils.gather_assignments(hashval_counts, [db])
    counts = count_lca_for_assignments(assignments, hashval_counts)
    print(counts)

    assert len(counts) == 3
    assert counts[lin] == 5  # makes sense b/c hashval2
    assert counts[lin2] == 0  # a;b;d (lin2) + a;b;d;e (lin3) -->a;b;d;e (lin3) only
    assert counts[lin3] == 3  # hashval3

    lca_lin = lca_utils.make_lineage("a;b")
    assert counts[lca_lin] == 2  # yes, b/c hashval


def test_count_lca_for_assignments_abund_5():
    # test basic mechanics of gather_assignments function with two lineages
    # and two hashvals when linages match but one has lower taxo detail
    hashval = 12345678
    hashval2 = 87654321
    hashval_counts = {}
    hashval_counts[hashval] = 2
    hashval_counts[hashval2] = 5

    lin = lca_utils.make_lineage("a;b;d")
    lin2 = lca_utils.make_lineage("a;b;d;e")

    db = FakeLCA_Database()
    db._set_lineage_assignment(hashval, set([lin, lin2]))
    db._set_lineage_assignment(hashval2, set([lin]))

    assignments = lca_utils.gather_assignments(hashval_counts, [db])
    counts = count_lca_for_assignments(assignments, hashval_counts)
    print(counts)

    assert len(counts) == 2
    assert counts[lin] == 5  # makes sense
    assert counts[lin2] == 2  # lin+lin2 yield just lin2


def test_is_lineage_match_1():
    # basic behavior: match at order and above, but not at family or below.
    lin1 = make_lineage("d__a;p__b;c__c;o__d;f__e")
    lin2 = make_lineage("d__a;p__b;c__c;o__d;f__f")

    assert is_lineage_match(lin1, lin2, "superkingdom")
    assert is_lineage_match(lin1, lin2, "phylum")
    assert is_lineage_match(lin1, lin2, "class")
    assert is_lineage_match(lin1, lin2, "order")
    assert not is_lineage_match(lin1, lin2, "family")
    assert not is_lineage_match(lin1, lin2, "genus")
    assert not is_lineage_match(lin1, lin2, "species")


def test_is_lineage_match_2():
    # match at family, and above, levels; no genus or species to match
    lin1 = make_lineage("d__a;p__b;c__c;o__d;f__f")
    lin2 = make_lineage("d__a;p__b;c__c;o__d;f__f")

    assert is_lineage_match(lin1, lin2, "superkingdom")
    assert is_lineage_match(lin1, lin2, "phylum")
    assert is_lineage_match(lin1, lin2, "class")
    assert is_lineage_match(lin1, lin2, "order")
    assert is_lineage_match(lin1, lin2, "family")
    assert not is_lineage_match(lin1, lin2, "genus")
    assert not is_lineage_match(lin1, lin2, "species")


def test_is_lineage_match_3():
    # one lineage is empty
    lin1 = make_lineage("")
    lin2 = make_lineage("d__a;p__b;c__c;o__d;f__f")

    assert not is_lineage_match(lin1, lin2, "superkingdom")
    assert not is_lineage_match(lin1, lin2, "family")
    assert not is_lineage_match(lin1, lin2, "order")
    assert not is_lineage_match(lin1, lin2, "class")
    assert not is_lineage_match(lin1, lin2, "phylum")
    assert not is_lineage_match(lin1, lin2, "genus")
    assert not is_lineage_match(lin1, lin2, "species")


def test_pop_to_rank_1():
    # basic behavior - pop to order?
    lin1 = make_lineage("d__a;p__b;c__c;o__d")
    lin2 = make_lineage("d__a;p__b;c__c;o__d;f__f")

    print(lin1)
    print(pop_to_rank(lin2, "order"))
    assert pop_to_rank(lin2, "order") == lin1


def test_pop_to_rank_2():
    # what if we're already above rank?
    lin2 = make_lineage("d__a;p__b;c__c;o__d;f__f")

    print(pop_to_rank(lin2, "species"))
    assert pop_to_rank(lin2, "species") == lin2



#! /usr/bin/env python
import sys
import argparse
import tomllib
import pprint
import json


def main():
    p = argparse.ArgumentParser()
    p.add_argument("toml_file")
    p.add_argument("-o", "--output-json", required=True)
    args = p.parse_args()

    with open(args.toml_file, "rb") as fp:
        d = tomllib.load(fp)

    # pprint.pprint(d)
    out_d = {}

    creators = []
    for author_d in d["project"]["authors"]:
        creators.append(author_d)

    out_d["creators"] = creators

    with open(args.output_json, "w") as fp:
        json.dump(out_d, fp)


if __name__ == "__main__":
    sys.exit(main())



"""concatenate signature files"""

usage = """

### `sourmash signature cat` - concatenate multiple signatures together

Concatenate signature files.

For example,

sourmash signature cat file1.sig file2.sig -o all.sig

will combine all signatures in `file1.sig` and `file2.sig` and put them
in the file `all.sig`.

"""

from sourmash.cli.utils import (
    add_moltype_args,
    add_ksize_arg,
    add_picklist_args,
    add_pattern_args,
)


def subparser(subparsers):
    # working on this
    subparser = subparsers.add_parser("cat", description=__doc__, usage=usage)
    subparser.add_argument("signatures", nargs="*")
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-d", "--debug", action="store_true", help="provide debugging output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="FILE",
        default="-",
        help="output signature to this file (default stdout)",
    )
    subparser.add_argument(
        "-u",
        "--unique",
        action="store_true",
        help="keep only distinct signatures, removing duplicates (based on md5sum)",
    )
    subparser.add_argument(
        "-f", "--force", action="store_true", help="try to load all files as signatures"
    )
    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_pattern_args(subparser)
    add_picklist_args(subparser)


def main(args):
    import sourmash

    return sourmash.sig.__main__.cat(args)



"""create a similarity matrix comparing many samples"""

usage = """

The `compare` subcommand compares one or more signatures (created with
`sketch`) using estimated Jaccard index [1] or (if signatures are
created with `-p abund`) the angular similarity [2]).

The default output is a text display of a similarity matrix where each
entry `[i, j]` contains the estimated Jaccard index between input
signature `i` and input signature `j`.  The output matrix can be saved
to a file with `--output <outfile.mat>` and used with the `sourmash
plot` subcommand (or loaded with `numpy.load(...)`.  Using `--csv
<outfile.csv>` will output a CSV file that can be loaded into other
languages than Python, such as R.

Command line usage:
```
sourmash compare file1.sig [ file2.sig ... ]
```

**Note:** compare by default produces a symmetric similarity matrix that can be used as an input to clustering. With `--containment`, however, this matrix is no longer symmetric and cannot formally be used for clustering.

[1] https://en.wikipedia.org/wiki/Jaccard_index
[2] https://en.wikipedia.org/wiki/Cosine_similarity#Angular_distance_and_similarity

---
"""

from sourmash.cli.utils import (
    add_ksize_arg,
    add_moltype_args,
    add_picklist_args,
    add_pattern_args,
    add_scaled_arg,
)


def subparser(subparsers):
    subparser = subparsers.add_parser("compare", description=__doc__, usage=usage)
    subparser.add_argument(
        "signatures", nargs="*", help="list of signatures to compare", default=[]
    )
    subparser.add_argument(
        "-q", "--quiet", action="store_true", help="suppress non-error output"
    )
    subparser.add_argument(
        "-o",
        "--output",
        metavar="F",
        help="file to which output will be written; default is terminal "
        "(standard output)",
    )
    subparser.add_argument(
        "--ignore-abundance",
        action="store_true",
        help="do NOT use k-mer abundances even if present",
    )
    subparser.add_argument(
        "--containment",
        action="store_true",
        help="calculate containment instead of similarity",
    )
    subparser.add_argument(
        "--max-containment",
        action="store_true",
        help="calculate max containment instead of similarity",
    )
    subparser.add_argument(
        "--avg-containment",
        "--average-containment",
        action="store_true",
        help="calculate average containment instead of similarity",
    )
    subparser.add_argument(
        "--estimate-ani",
        "--ANI",
        "--ani",
        action="store_true",
        help="return ANI estimated from jaccard, containment, average containment, or max containment; see https://doi.org/10.1101/2022.01.11.475870",
    )
    subparser.add_argument(
        "--from-file",
        help="a text file containing a list of files to load signatures from",
    )
    subparser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="continue past errors in file loading",
    )
    subparser.add_argument(
        "--csv",
        metavar="F",
        help="write matrix to specified file in CSV format (with column headers)",
    )
    subparser.add_argument(
        "--labels-to",
        "--labels-save",
        help="a CSV file containing label information",
    )
    subparser.add_argument(
        "-p",
        "--processes",
        metavar="N",
        type=int,
        default=None,
        help="Number of processes to use to calculate similarity",
    )
    subparser.add_argument(
        "--distance-matrix",
        action="store_true",
        help="output a distance matrix, instead of a similarity matrix",
    )
    subparser.add_argument(
        "--similarity-matrix",
        action="store_false",
        dest="distance_matrix",
        help="output a similarity matrix; this is the default",
    )

    add_ksize_arg(subparser)
    add_moltype_args(subparser)
    add_picklist_args(subparser)
    add_pattern_args(subparser)
    add_scaled_arg(subparser)


def main(args):
    import sourmash

    return sourmash.commands.compare(args)



"""
Tests for the 'sourmash signature grep' command line.
"""

import shutil
import os
import csv
import gzip

import pytest

import sourmash_tst_utils as utils
import sourmash
from sourmash_tst_utils import SourmashCommandFailed
from sourmash.signature import load_signatures_from_json, save_signatures_to_json

## command line tests


def test_grep_1_sig_name(runtmp):
    # search on substring in name
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "Shewanella", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella" in ss.name
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_1_sig_name_case_sensitive(runtmp):
    # search on substring in name
    sig47 = utils.get_test_data("47.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash("sig", "grep", "shewanella", sig47)


def test_grep_1_sig_name_case_insensitive(runtmp):
    # search on substring in name, case insensitive
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "-i", "shewanella", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella" in ss.name
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_1_sig_name_exclude(runtmp):
    # search on substring in name, case insensitive
    sig47 = utils.get_test_data("47.fa.sig")

    # no matches!
    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash("sig", "grep", "-v", "Shewanella", sig47)


def test_grep_2_sig_md5(runtmp):
    # search on substring in md5
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "ce52952152f0", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_2_sig_md5_case_sensitive(runtmp):
    # case sensitive no match
    sig47 = utils.get_test_data("47.fa.sig")

    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash("sig", "grep", "CE52952152f0", sig47)


def test_grep_2_sig_md5_case_insensitive(runtmp):
    # search on substring in md5, case insensitive
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "-i", "CE52952152f0", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_3_filename(runtmp):
    # filename match
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "47.fa", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "47.fa" in ss.filename
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_3_filename_regexp(runtmp):
    # search for a regexp on filename
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.run_sourmash("sig", "grep", "^47.fa", sig47)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "7.fa" in ss.filename
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"


def test_grep_4_no_manifest(runtmp):
    # fail search when no manifest, by default
    sbt = utils.get_test_data("v6.sbt.zip")

    with pytest.raises(SourmashCommandFailed):
        runtmp.run_sourmash("sig", "grep", "e60265", sbt)

    print(runtmp.last_result.err)
    assert "ERROR on filename" in runtmp.last_result.err
    assert (
        "sig grep requires a manifest by default, but no manifest present."
        in runtmp.last_result.err
    )


def test_grep_4_no_manifest_ok(runtmp):
    # generate manifest if --no-require-manifest
    sbt = utils.get_test_data("v6.sbt.zip")

    runtmp.run_sourmash("sig", "grep", "e60265", sbt, "--no-require-manifest")

    ss = load_signatures_from_json(runtmp.last_result.out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "e60265" in ss.md5sum()


def test_grep_5_zip_include(runtmp):
    # search zip, include on case sensitive match to name
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash("sig", "grep", "--dna", "OS223", allzip)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_5_zip_include_picklist(runtmp):
    # search zip, include on case sensitive match to name
    allzip = utils.get_test_data("prot/all.zip")

    pickfile = runtmp.output("pick.csv")
    with open(pickfile, "w", newline="") as fp:
        w = csv.DictWriter(fp, fieldnames=["md5"])
        w.writeheader()
        w.writerow(dict(md5="09a08691ce52952152f0e866a59f6261"))
        w.writerow(dict(md5="38729c6374925585db28916b82a6f513"))

    runtmp.run_sourmash(
        "sig", "grep", "--dna", "OS223", allzip, "--picklist", f"{pickfile}:md5:md5"
    )

    out = runtmp.last_result.out
    print(out)
    err = runtmp.last_result.err
    print(err)
    assert "for given picklist, found 2 matches to 2 distinct values" in err

    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_5_zip_include_case_insensitive(runtmp):
    # search zip, include on case insensitive match to name
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash("sig", "grep", "--dna", "-i", "os223", allzip)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_5_zip_exclude(runtmp):
    # search zip, exclude on case-sensitive match
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash("sig", "grep", "--dna", "-v", "OS185", allzip)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_5_zip_exclude_case_insensitive(runtmp):
    # search zip, exclude on case-insensitive match
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash("sig", "grep", "--dna", "-vi", "os185", allzip)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_6_zip_manifest_csv(runtmp):
    # do --csv and use result as picklist
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash("sig", "grep", "--dna", "OS223", allzip, "--csv", "match.csv")

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"

    # now run cat with picklist
    runtmp.run_sourmash("sig", "cat", allzip, "--picklist", "match.csv::manifest")

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_grep_6_zip_manifest_csv_gz(runtmp):
    # do --csv and use result as picklist
    allzip = utils.get_test_data("prot/all.zip")

    runtmp.run_sourmash(
        "sig", "grep", "--dna", "OS223", allzip, "--csv", "match.csv.gz"
    )

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"

    # check that match.csv.gz is a gzip file
    with gzip.open(runtmp.output("match.csv.gz"), "rt", newline="") as fp:
        fp.read()

    # now run cat with picklist
    runtmp.run_sourmash("sig", "cat", allzip, "--picklist", "match.csv.gz::manifest")

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella baltica OS223" in ss.name
    assert ss.md5sum() == "38729c6374925585db28916b82a6f513"


def test_sig_grep_7_lca(runtmp):
    # extract 47 from an LCA database, with --no-require-manifest
    allzip = utils.get_test_data("lca/47+63.lca.json")
    sig47 = utils.get_test_data("47.fa.sig")

    runtmp.sourmash(
        "sig",
        "grep",
        "50a9274021e4",
        allzip,
        "--no-require-manifest",
        "-o",
        "matches.sig",
    )

    match = sourmash.load_file_as_signatures(runtmp.output("matches.sig"))
    match = list(match)[0]

    ss47 = sourmash.load_file_as_signatures(sig47)
    ss47 = list(ss47)[0]

    ss47 = ss47.to_mutable()
    ss47.minhash = ss47.minhash.downsample(scaled=10000)

    assert ss47.minhash == match.minhash


def test_sig_grep_7_picklist_md5_lca_fail(runtmp):
    # extract 47 from an LCA database, using a picklist w/full md5 => fail
    allzip = utils.get_test_data("lca/47+63.lca.json")

    # select on any of these attributes
    row = dict(
        exactName="NC_009665.1 Shewanella baltica OS185, complete genome",
        md5full="50a9274021e43eda8b2e77f8fa60ae8e",
        md5short="50a9274021e43eda8b2e77f8fa60ae8e"[:8],
        fullIdent="NC_009665.1",
        nodotIdent="NC_009665",
    )

    # make picklist
    picklist_csv = runtmp.output("pick.csv")
    with open(picklist_csv, "w", newline="") as csvfp:
        w = csv.DictWriter(csvfp, fieldnames=row.keys())
        w.writeheader()
        w.writerow(row)

    picklist_arg = f"{picklist_csv}:md5full:md5"
    with pytest.raises(SourmashCommandFailed):
        runtmp.sourmash(
            "sig",
            "grep",
            "50a92740",
            allzip,
            "--picklist",
            picklist_arg,
            "--no-require-manifest",
        )

    # this happens b/c the implementation of 'grep' uses picklists, and
    # LCA databases don't support multiple picklists.
    print(runtmp.last_result.err)
    assert (
        "This input collection doesn't support 'grep' with picklists."
        in runtmp.last_result.err
    )


def test_sig_grep_8_count(runtmp):
    zips = [
        "prot/all.zip",
        "prot/dayhoff.sbt.zip",
        "prot/dayhoff.zip",
        "prot/hp.sbt.zip",
        "prot/hp.zip",
        "prot/protein.sbt.zip",
        "prot/protein.zip",
    ]

    zip_src = [utils.get_test_data(x) for x in zips]

    os.mkdir(runtmp.output("prot"))
    for src, dest in zip(zip_src, zips):
        shutil.copyfile(src, runtmp.output(dest))

    runtmp.sourmash("sig", "grep", "-c", "0015939", *zips)

    out = runtmp.last_result.out
    err = runtmp.last_result.err

    print(out)
    print(err)

    assert "(no signatures will be saved because of --silent/--count)." in err

    for line in """\
6 matches: prot/all.zip
2 matches: prot/dayhoff.sbt.zip
2 matches: prot/dayhoff.zip
2 matches: prot/hp.sbt.zip
2 matches: prot/hp.zip
2 matches: prot/protein.sbt.zip
2 matches: prot/protein.zip
""".splitlines():
        assert line.strip() in out


def test_sig_grep_identical_md5s(runtmp):
    # test that we properly handle different signatures with identical md5s
    sig47 = utils.get_test_data("47.fa.sig")
    ss = load_signatures_from_json(sig47)
    sig = list(ss)[0]
    new_sig = sig.to_mutable()
    new_sig.name = "foo"
    sig47foo = runtmp.output("foo.sig")
    # this was only a problem when the signatures are stored in the same file
    with open(sig47foo, "w") as fp:
        save_signatures_to_json([new_sig, sig], fp)

    runtmp.run_sourmash("sig", "grep", "-i", "foo", sig47foo)

    out = runtmp.last_result.out
    ss = load_signatures_from_json(out)
    ss = list(ss)
    assert len(ss) == 1
    ss = ss[0]
    assert "Shewanella" not in ss.name
    assert "foo" in ss.name
    assert ss.md5sum() == "09a08691ce52952152f0e866a59f6261"
